/**
 * Contain the logic for the prova bank modal.
 *
 * @module     mod_quiz/modal_prova_bank
 * @package    mod_quiz
 */
define([
    'jquery',
    'core/yui',
    'core/notification',
    'core/modal',
    'core/modal_events',
    'core/modal_registry',
    'core/fragment'
],
function(
    $,
    Y,
    Notification,
    Modal,
    ModalEvents,
    ModalRegistry,
    Fragment
) {

    var registered = false;
    var SELECTORS = {
        SELECT_ROW: '.bankprovatablerow',
        POS_SELECT_ROW_RADIO: 'td input[name="provaid"]',
        SELECT_ROW_RADIO: '.bankprovatablerow td input[name="provaid"]',
        SELECTED_ROW_RADIO: '.bankprovatablerow td input[name="provaid"]:checked',
        ANCHOR: 'a[href]',
        PREVIEW_CONTAINER: '.previewaction',
        SEARCH_OPTIONS: '#advancedsearch',
        DISPLAY_OPTIONS: '#displayoptions',
        ADD_PROVA_FORM: 'form.bankprovasubmitform',
        // *SEOXEdit 2019092400 '1 line'
        BUSCAPROVAS_HEADER: 'form[data="buscaprovas"]',
        BUSCAPROVAS: '[data="buscaprovas"]',
    };

    /**
     * Constructor for the Modal.
     *
     * @param {object} root The root jQuery element for the modal
     */
    var ModalProvaBank = function(root) {
        Modal.call(this, root);

        this.contextId = null;
        this.pageurl = null;
    };

    ModalProvaBank.TYPE = 'mod_prova-prova-bank';
    ModalProvaBank.prototype = Object.create(Modal.prototype);
    ModalProvaBank.prototype.constructor = ModalProvaBank;

    /**
     * Save the Moodle context id that the question bank is being
     * rendered in.
     *
     * @method setContextId
     * @param {int} id
     */
    ModalProvaBank.prototype.setContextId = function(id) {
        this.contextId = id;
    };

    /**
     * Retrieve the saved Moodle context id.
     *
     * @method getContextId
     * @return {int}
     */
    ModalProvaBank.prototype.getContextId = function() {
        return this.contextId;
    };

    /**
     * Save the current pageurl of the page bank.
     *
     * @method setContextId
     * @param {int} id
     */
    ModalProvaBank.prototype.setPageurl = function(id) {
        this.pageurl = id;
    };

    /**
     * Retrieve the saved Moodle pageurl.
     *
     * @method getContextId
     * @return {int}
     */
    ModalProvaBank.prototype.getPageurl = function() {
        return this.pageurl;
    };


    /**
     * Override the parent show function.
     *
     * Reload the body contents when the modal is shown. The current
     * window URL is used to inform the new content that should be
     * displayed.
     *
     * @method show
     * @return {void}
     */
    ModalProvaBank.prototype.show = function() {
        this.reloadBodyContent(window.location.search);
        return Modal.prototype.show.call(this);
    };

    /**
     * Replaces the current body contents with a new version of the question
     * bank.
     *
     * The contents of the question bank are generated using the provided
     * query string.
     *
     * @method reloadBodyContent
     * @param {string} queryString URL encoded string.
     */
    ModalProvaBank.prototype.reloadBodyContent = function(queryString) {
        // Load the question bank fragment to be displayed in the modal.
        var promise = Fragment.loadFragment(
            'mod_prova',
            'prova_bank',
            this.getContextId(),
            {
                querystring: queryString
            }
        ).fail(Notification.exception);

        this.setBody(promise);
    };

    /**
     * Update the URL of the anchor element that the user clicked on to make
     * sure that the question is added to the correct page.
     *
     * @method handleRowClick
     * @param {event} e A JavaScript event
     * @param {object} anchorElement The anchor element that was triggered
     */
    ModalProvaBank.prototype.handleRowClick = function(e, anchorElement) {
        $(SELECTORS.SELECT_ROW_RADIO).prop('checked', false);
        $(SELECTORS.SELECT_ROW).removeAttr('checked');
        anchorElement.find(SELECTORS.POS_SELECT_ROW_RADIO).prop('checked', true);
        anchorElement.attr('checked','checked');
    };

    /**
     * Open a popup window to show the preview of the question.
     *
     * @method handlePreviewContainerEvent
     * @param {event} e A JavaScript event
     * @param {object} anchorElement The anchor element that was triggered
     */
    ModalProvaBank.prototype.handlePreviewContainerEvent = function(e, anchorElement) {
        var popupOptions = [
            'height=700',
            'width=800',
            'top=0',
            'left=0',
            'menubar=0',
            'location=0',
            'scrollbars',
            'resizable',
            'toolbar',
            'status',
            'directories=0',
            'fullscreen=0',
            'dependent'
        ];
        window.openpopup(e, {
            url: anchorElement.attr('href'),
            name: 'questionpreview',
            options: popupOptions.join(',')
        });
    };

    /**
     * Reload the modal body with the new display options the user has selected.
     *
     * A query string is built using the form elements to be used to generate the
     * new body content.
     *
     * @method handleDisplayOptionFormEvent
     * @param {event} e A JavaScript event
     */
    ModalProvaBank.prototype.handleDisplayOptionFormEvent = function(e) {
        // Stop propagation to prevent other wild event handlers
        // from submitting the form on change.
        e.stopPropagation();
        e.preventDefault();

        var form = $(e.target).closest(SELECTORS.DISPLAY_OPTIONS);
        var queryString = '?' + form.serialize();
        this.reloadBodyContent(queryString);
    };

    /**
     * Listen for changes to the display options form.
     *
     * This handles the user changing:
     *      - The quiz category select box
     *      - The tags to filter by
     *      - Show/hide questions from sub categories
     *      - Show/hide old questions
     *
     * @method registerDisplayOptionListeners
     */
    ModalProvaBank.prototype.registerDisplayOptionListeners = function() {
        // Listen for changes to the display options form.
        this.getModal().on('change', SELECTORS.DISPLAY_OPTIONS, function(e) {
            // Get the element that was changed.
            var modifiedElement = $(e.target);
            if (modifiedElement.attr('aria-autocomplete')) {
                // If the element that was change is the autocomplete
                // input then we should ignore it because that is for
                // display purposes only.
                return;
            }

            this.handleDisplayOptionFormEvent(e);
        }.bind(this));

        // Listen for the display options form submission because the tags
        // filter will submit the form when it is changed.
        this.getModal().on('submit', SELECTORS.DISPLAY_OPTIONS, function(e) {
            this.handleDisplayOptionFormEvent(e);
        }.bind(this));
    };

    /**
     * Prepera o query do filtro de busca
     */
    ModalProvaBank.prototype.prepareBuscaFilters = function(anchorElement) {
        const inputs = $(anchorElement).find(SELECTORS.BUSCAPROVAS);
        var params = [];
        var key = null;
        var value = null;
        inputs.each(function(index, elem) {
            key = (elem.id).replace('id_', '');
            key = (key).replace('_random', '');
            value = $(elem).val();
            params[key] = (Array.isArray(value)) ? value.join('-') : value;
        });
        var queryString = Object.keys(params).map(function(key) {
            return key + '=' + params[key]
        }).join('&');
        return queryString;
    };

    /**
     * Set up all of the event handling for the modal.
     *
     * @method registerEventListeners
     */
    ModalProvaBank.prototype.registerEventListeners = function() {
        // Apply parent event listeners.
        Modal.prototype.registerEventListeners.call(this);

        // Set up the event handlers for all of the display options.
        this.registerDisplayOptionListeners();

        // Redireciona o submit para url correta
        this.getModal().on('submit', SELECTORS.ADD_PROVA_FORM, function(e) {
            e.preventDefault();
            var provaid = $(SELECTORS.SELECTED_ROW_RADIO).val()
            const currentPath = window.location.pathname + window.location.search;
            window.location.href = currentPath + (provaid ? '&' + $.param({'provaid': provaid}) : '');

        }.bind(this));

        // Pinta a celulas das linhas
        this.getModal().on('click', SELECTORS.SELECT_ROW, function(e) {
            var anchorElement = $(e.currentTarget);
            this.handleRowClick(e, anchorElement);
        }.bind(this));

        // Libera o botão de prwview para ser clicado
        this.getModal().on('click', SELECTORS.PREVIEW_CONTAINER, function(e) {
            e.stopPropagation();
            return false;
        }.bind(this));

        this.getModal().on('click', SELECTORS.ANCHOR, function(e) {
            var anchorElement = $(e.currentTarget);


            // If the anchor element was a preview question link.
            if (anchorElement.closest(SELECTORS.PREVIEW_CONTAINER).length) {
                this.handlePreviewContainerEvent(e, anchorElement);
                return;
            }

            // Click on expand/collaspse search-options. Has its own handler.
            // We should not interfere.
            if (anchorElement.closest(SELECTORS.SEARCH_OPTIONS).length) {
                return;
            }

            // Anything else means reload the pop-up contents.
            e.preventDefault();
            this.reloadBodyContent(anchorElement.prop('search') + '&' + this.prepareBuscaFilters($(SELECTORS.BUSCAPROVAS_HEADER)));
        }.bind(this));

        // Disable the form change checker when the body is rendered.
        this.getRoot().on(ModalEvents.bodyRendered, function() {
            // Make sure the form change checker is disabled otherwise it'll
            // stop the user from navigating away from the page once the modal
            // is hidden.
            Y.use('moodle-core-formchangechecker', function() {
                M.core_formchangechecker.reset_form_dirty_state();
            });
        });

        this.getModal().on('submit', SELECTORS.BUSCAPROVAS_HEADER, function(e) { 
            e.preventDefault();

            this.reloadBodyContent(window.location.search + '&' + this.prepareBuscaFilters($(e.currentTarget)));
        }.bind(this));

    };

    // Automatically register with the modal registry the first time this module is
    // imported so that you can create modals of this type using the modal factory.
    if (!registered) {
        ModalRegistry.register(
            ModalProvaBank.TYPE,
            ModalProvaBank,
            'core/modal'
        );

        registered = true;
    }

    return ModalProvaBank;
});
