<?php
/**
 * Define a tela de banco de provas do cadastro de aplicação de prova
 *
 * @package   provacomp_addprovas
 */

namespace mod_prova\prova;
defined('MOODLE_INTERNAL') || die();

/**
 * Classe responsável por criar a listagem de provas para adicionar na aplicação
 */
class bank_view {

	/** @var \moodle_url the current page url. */
	protected $pageurl = false;

	/** @var \array parametros de busca. */
	protected $search = null;

	/** @var \str str de busca de provas no banco */
	protected $loadsql = null;

	/** @var \str sql que consulta quantidade de provas */
	protected $countsql = null;

	/** @var \str all url important params **/
	protected $urlparams = null;

	public function __construct($pageurl, $pagevars = null) {
		\mod_prova\manager::has_permission_to_access();

		$this->$pageurl = $pageurl;

		if ($pagevars) {
			$this->prepare_global_attributes($pagevars);
		}
	}

	protected function prepare_global_attributes($pagevars) {
		$str_targets = array('md_areas', 'md_subareas', 'md_assuntos', 'md_text');
        $array_targets = array('md_dificuldades');

        $search = array();

        foreach ($pagevars as $key => $value) {
            if (in_array($key, $str_targets)) {
                $search[$key] = $value;
                unset($pagevars[$key]);
            } elseif (in_array($key, $array_targets) and $value !== '') {
                $search[$key] = explode('-', $value);
                unset($pagevars[$key]);
            }
        }


        $urlparams = '';

        foreach ($pagevars as $key => $value) {
        	$urlparams .= \html_writer::empty_tag('input', array('type' => 'hidden', 'name' => $key, 'value' => $value));
        }

		$this->search = $search;
		$this->urlparams = $urlparams;
	}

	protected function get_provas_count() {
		global $DB;
		return $DB->count_records_sql($this->countsql, []);
	}

	protected function load_page_provas($page, $perpage) {
		global $DB;
		$provas = $DB->get_recordset_sql($this->loadsql, null, $page * $perpage, $perpage);
		if (!$provas->valid()) {
			// No provas on this page. Reset to page 0.
			$provas->close();
			$provas = $DB->get_recordset_sql($this->loadsql, null, 0, $perpage);
		}
		return $provas;
	}

	/**
	 * Prepara um objeto com o conteúdo do da tabela
	 *
	 * @param \int $page. Página de provas específica
	 * @param \int $perpage. Quantidade de provas por página
	 */
	public function render($page, $perpage) {

		ob_start();
		$this->display($page, $perpage);
		$out = ob_get_contents();
		ob_end_clean();
		return $out;
	}

	/**
	 * Gerencia o que mostrar na tabela
	 *
	 * @param \int $page. Página de provas específica
	 * @param \int $perpage. Quantidade de provas por página
	 */
	public function display($page, $perpage) {
		global $PAGE, $CFG;

		$this->display_prova_bank_header();

		// Mostra formulário de pesquisa
		$this->display_buscaprovas_form();

		// // Continues with list of questions.
		$this->display_prova_list($page, $perpage);
	}

	/**
	 * Display the header element for the prova bank.
	 */
	protected function display_prova_bank_header() {
		global $OUTPUT;
		echo $OUTPUT->heading(get_string('provabank', 'prova'), 2);
	}

	/**
	 * Mostra formulário de busca no modal
	 */
	protected function display_buscaprovas_form() {
		global $PAGE;

		$advancedsearch = 'advancedsearch';
		$thisform = new \provacomp_buscaprovas\form\buscaprovas_form();
        $thisform->add_classes_to_elements();
		$params = $thisform->get_filtros_areas_content();

		print_collapsible_region_start('', $advancedsearch, get_string('advancedsearchoptions', 'question'), 'question_bank_advanced_search');

		if (!empty($this->search)) {
			$thisform->populate_elements($this->search);
		}
		$thisform->display();

		$PAGE->requires->yui_module(
			'moodle-provacomp_buscaquestoes-filtro_searchquestions',
			'M.provacomp_buscaquestoes.filtro_searchquestions.init',
			$params
		);

		print_collapsible_region_end();
	}

	/**
	 * Gera a lista de provas com base na busca
	 *
	 * @param \int $page. Página de provas específica
	 * @param \int $perpage. Quantidade de provas por página
	 */
	protected function display_prova_list($page, $perpage) {
		global $OUTPUT;

		// This function can be moderately slow with large question counts and may time out.
		// We probably do not want to raise it to unlimited, so randomly picking 5 minutes.
		// Note: We do not call this in the loop because quiz ob_ captures this function (see raise() PHP doc).
		\core_php_time_limit::raise(300);

		$this->build_query();

		// Get provas
		$totalnumber = $this->get_provas_count();
		if ($totalnumber == 0) {
			return;
		}
		$provas = $this->load_page_provas($page, $perpage);

		echo '<div class="categorypagingbarcontainer">';
		$pagingbar = new \paging_bar($totalnumber, $page, $perpage, $this->pageurl);
		$pagingbar->pagevar = 'qpage';
		echo $OUTPUT->render($pagingbar);
		echo '</div>';

		echo \html_writer::start_tag('form', array('class' => 'bankprovasubmitform', 'action' => 'modedit.php', 'method' => 'GET'));
		echo \html_writer::empty_tag('input', array('type' => 'hidden', 'name' => 'sesskey', 'value' => sesskey()));
		echo $this->urlparams;
		echo \html_writer::start_tag('div', array('class' => 'bankprovacontainer', 'id' => 'provascontainer'));

		$table = new \html_table();
		$table->attributes = ['class' => 'generaltable bankprovatable'];
		$table->head = array('Nome', 'Áreas', 'Dificuldade', 'Vezes Utilizada', 'Qtd de Questões', 'Ações', '');

		foreach ($provas as $prova) {
			$row = new \html_table_row();
			$row->attributes = ['class' => 'bankprovatablerow'];
			$row->cells[] = $prova->nome;
			$row->cells[] = $this->prepare_areas($prova->area, $prova->subarea, $prova->assunto);
			// TODO: (5) configura dificuldade corretamente
			$row->cells[] = \provacomp_buscaprovas\form\buscaprovas_form::$dificuldadeslist[$prova->dificuldade];
			$row->cells[] = intval($prova->vezes_usada);
			$row->cells[] = intval($prova->nro_questions);
			$row->cells[] = $this->prepare_acoes($prova->provaid);
			$row->cells[] = \html_writer::tag('input', '', array('type' => 'radio', 'value' => $prova->provaid, 'name' => 'provaid'));
			$table->data[] = $row;
		}
		$provas->close();
		echo \html_writer::table($table);

		$this->display_bottom_controls($totalnumber, $recurse, $category, $catcontext, $addcontexts);

		echo \html_writer::end_tag('div');
		echo \html_writer::end_tag('form');

		echo '<div class="categorypagingbarcontainer pagingbottom">';
		echo $OUTPUT->render($pagingbar);
		echo '</div>';
	}

	protected function build_query() {
		global $DB;

		$selects = 'SELECT
			prova.id AS provaid,
			prova.name AS nome,
			area.name AS area,
			subarea.name AS subarea,
			assunto.name AS assunto,
			prova.dificuldade AS dificuldade,
			prova.vezes_usada AS vezes_usada,
			COUNT(slots.id) AS nro_questions';

		$from = ' FROM {seox_prova} prova';
		$join = ' LEFT JOIN {seox_prova_slots} slots
			ON prova.id = slots.provaid
			LEFT JOIN {seox_area} area
			ON prova.id_area = area.id
			LEFT JOIN {seox_subarea} subarea
			ON prova.id_subarea = subarea.id
			LEFT JOIN mdl_seox_assunto assunto
			ON prova.id_assunto = assunto.id';

		$filters = (!empty($this->search)) ? $this->build_search_query() : [];
		$noemptyprova = ['EXISTS (SELECT * FROM {seox_prova_slots} WHERE provaid = prova.id)'];
		$query = array_merge($filters, $noemptyprova);
		$where = " WHERE " . implode(' AND ', $query);
		$groupby = ' GROUP BY prova.id, area.id, subarea.id, assunto.id';

		$this->countsql = 'SELECT count(1)' . $from . $where;
		$this->loadsql = $selects . $from . $join . $where . $groupby . ' ORDER BY prova.id DESC';
	}

	protected function build_search_query() {
		$search = $this->search;
        $data = array();

        if ($search['md_areas'] != 0) {
            $data[] = 'prova.id_area = '.$search['md_areas'];
        }
        if ($search['md_subareas'] != 0) {
            $data[] = 'prova.id_subarea = '.$search['md_subareas'];
        }
        if ($search['md_assuntos'] != 0) {
            $data[] = 'prova.id_assunto = '.$search['md_assuntos'];
        }
        if (!empty($search['md_dificuldades'])) {
            $str = 'prova.dificuldade IN (';
            $str .= implode(',', $search['md_dificuldades']);
            $str .= ')';
            $data[] = $str;
        }
        if ($search['md_text'] and $search['md_text'] != '') {
            $data[] = "prova.name LIKE '%".$search['md_text']."%'";
        }

        return $data;
    } 

	protected function prepare_areas($area, $subarea, $assunto) {
		return \html_writer::start_div('bankprovahideareas') .
		\html_writer::tag('span', $area . ' (...)') .
		\html_writer::tag('p', $area) .
		\html_writer::tag('p', $subarea) .
		\html_writer::tag('p', $assunto) .
		\html_writer::end_div();
	}

	protected function prepare_acoes($provaid) {
		global $OUTPUT;

		$previewurl = new \moodle_url('/mod/prova/components/addprovas/preview.php', array(
			'provaid' => $provaid,
		));
		$previewimage = $OUTPUT->pix_icon('t/preview', get_string('preview'), '', array('class' => 'iconsmall'));
		$previewaction = new \popup_action('click', $previewurl, 'provapreview', array(
			'height' => 600,
			'width' => 800,
		));
		return $OUTPUT->action_link($previewurl, $previewimage, $previewaction, array('class' => 'previewaction'));
	}

	protected function display_bottom_controls() {

		echo '<div class="modulespecificbuttonscontainer">';

		$params = array(
			'type' => 'submit',
			'name' => 'add',
			'class' => 'btn btn-primary',
			'value' => get_string('addprovatoactivity', 'prova'),
		);
		echo \html_writer::empty_tag('input', $params);

		echo "</div>\n";
	}

}