<?php
/**
 * External question API
 *
 * @package    provacomp_addprovas
 * @category   external
 */

defined('MOODLE_INTERNAL') || die();

// require_once($CFG->libdir. "/externallib.php");
require_once($CFG->dirroot . '/question/engine/lib.php');
require_once($CFG->dirroot . '/question/engine/datalib.php');
require_once($CFG->libdir . '/questionlib.php');

/**
 * Question external functions
 *
 * @package    core_question
 * @category   external
 * @copyright  2016 Pau Ferrer <pau@moodle.com>
 * @license    http://www.gnu.org/copyleft/gpl.html GNU GPL v3 or later
 * @since Moodle 3.1
 */
class core_question_external extends external_api {

    /**
     * Returns description of method parameters.
     *
     * @return external_function_parameters.
     */
    public static function get_random_question_summaries_parameters() {
        return new external_function_parameters([
            'area' => new external_value(PARAM_INT, 'Área da questão'),
            'subarea' => new external_value(PARAM_INT, 'Subárea da questão'),
            'assunto' => new external_value(PARAM_INT, 'Assunto da questão'),
            'dificuldade' => new external_value(PARAM_INT, 'Dificuldade da questão'),
            'qtype' => new external_value(PARAM_ALPHA, 'Tipo da questão'),
            'text' => new external_value(PARAM_ALPHA, 'Um pedaço do texto da questão'),
            'limit' => new external_value(PARAM_INT, 'Maximum number of results to return', VALUE_DEFAULT, 0),
            'offset' => new external_value(PARAM_INT, 'Number of items to skip from the begging of the result set', VALUE_DEFAULT, 0)
        ]);
    }

    /**
     * Gets the list of random questions for the given criteria. The questions
     * will be exported in a summaries format and won't include all of the
     * question data.
     *
     * @param int $categoryid Category id to find random questions
     * @param bool $includesubcategories Include the subcategories in the search
     * @param int[] $tagids Only include questions with these tags
     * @param int $contextid The context id where the questions will be rendered
     * @param int $limit Maximum number of results to return
     * @param int $offset Number of items to skip from the beginning of the result set.
     * @return array The list of questions and total question count.
     */
    public static function get_random_question_summaries(
        $area,
        $subarea,
        $assunto,
        $dificuldade,
        $qtype,
        $text,
        $limit = 0,
        $offset = 0
    ) {
        global $DB, $PAGE, $CFG;
        require_once($CFG->dirroot . '/mod/prova/managelib.php');

        // Parameter validation.
        $params = self::validate_parameters(
            self::get_random_question_summaries_parameters(),
            [
                'area' => $area,
                'subarea' => $subarea,
                'assunto' => $assunto,
                'dificuldade' => $dificuldade,
                'qtype' => $qtype,
                'text' => $text,
                'limit' => $limit,
                'offset' => $offset
            ]
        );

        $search_query = array(
            'md_area' => $params['area'];
            'md_subarea' => $params['subarea'];
            'md_assunto' => $params['assunto'];
            'md_dificuldade' => $params['dificuldade'];
            'md_qtype' => $params['qtype'];
            'md_text' => $params['text'];
        );

        $limit = $params['limit'];
        $offset = $params['offset'];

        $contextid = context_course::instance(SEOX_COURSE_ID)->id;
        $context = \context::instance_by_id($contextid);

        $serach_query = implode(' AND ', \provacomp_addprovas\action\addprovas_actions::build_search_query($search_query));
        $questionids = $DB->get_fieldset_sql('SELECT id_question FROM {seox_questions} WHERE ' . $serach_query);

        $questions = $DB->get_records_list(
            'question',
            'id',
            $questionids,
            'id',
            '*',
            $offset,
            $limit
        );

        $totalcount = count($questions);
        $renderer = $PAGE->get_renderer('core');

        $formattedquestions = array_map(function($question) use ($context, $renderer) {
            $exporter = new \core_question\external\question_summary_exporter($question, ['context' => $context]);
            return $exporter->export($renderer);
        }, $questions);

        return [
            'totalcount' => $totalcount,
            'questions' => $formattedquestions
        ];
    }

    /**
     * Returns description of method result value.
     */
    public static function  get_random_question_summaries_returns() {
        return new external_single_structure([
            'totalcount' => new external_value(PARAM_INT, 'total number of questions in result set'),
            'questions' => new external_multiple_structure(
                \core_question\external\question_summary_exporter::get_read_structure()
            )
        ]);
    }
}
