<?php 
/**
 * Classe que simula a classe do plugin quiz Structure
 */

namespace provacomp_addprovas;
defined('MOODLE_INTERNAL') || die();

require_once($CFG->dirroot . '/mod/prova/components/addprovas/lib.php');


class structure {

	protected $provaobj = null;

	protected $course_params = null;

    /**
     * @var \stdClass[] the questions in this prova. Contains the row from the questions
     * table, with the data from the prova_slots table added, and also question_categories.contextid.
     */
    protected $questions = array();

    /** @var \stdClass[] prova_slots.slot => the prova_slots rows for this prova, agumented by sectionid. */
    protected $slotsinorder = array();

    /**
     * @var \stdClass[] currently a dummy. Holds data that will match the
     * prova_sections, once it exists.
     */
    protected $sections = array();

    /** @var bool caches the results of can_be_edited. */
    protected $canbeedited = null;

    /** @var bool caches the results of can_add_random_question. */
    protected $canaddrandom = null;

    /** @var bool tracks whether tags have been loaded */
    protected $hasloadedtags = false;

    /**
     * @var \stdClass[] the tags for slots. Indexed by slot id.
     */
    protected $slottags = array();

    /**
     * Create an instance of this class representing an empty prova.
     * @return structure
     */
    public static function create() {
        return new self();
    }

	public static function create_for_prova($provaobj) {
        $structure = new self();
        $structure->provaobj = $provaobj;
        $structure->populate_structure($provaobj);
        $structure->recalculate_difficulty();
        return $structure;
    }


    /**
     * Whether there are any questions in the prova.
     * @return bool true if there is at least one question in the prova.
     */
    public function has_questions() {
        return !empty($this->questions);
    }

    /**
     * Get the number of questions in the prova.
     * @return int the number of questions in the prova.
     */
    public function get_question_count() {
        return count($this->questions);
    }

    /**
     * Get the information about the question with this id.
     * @param int $questionid The question id.
     * @return \stdClass the data from the questions table, augmented with
     * question_category.contextid, and the prova_slots data for the question in this prova.
     */
    public function get_question_by_id($questionid) {
        return $this->questions[$questionid];
    }

    /**
     * Get the information about the question in a given slot.
     * @param int $slotnumber the index of the slot in question.
     * @return \stdClass the data from the questions table, augmented with
     * question_category.contextid, and the prova_slots data for the question in this prova.
     */
    public function get_question_in_slot($slotnumber) {
        return $this->questions[$this->slotsinorder[$slotnumber]->questionid];
    }

    /**
     * Get the displayed question number (or 'i') for a given slot.
     * @param int $slotnumber the index of the slot in question.
     * @return string the question number ot display for this slot.
     */
    public function get_displayed_number_for_slot($slotnumber) {
        return $this->slotsinorder[$slotnumber]->displayednumber;
    }

    /**
     * Get the page a given slot is on.
     * @param int $slotnumber the index of the slot in question.
     * @return int the page number of the page that slot is on.
     */
    public function get_page_number_for_slot($slotnumber) {
        return $this->slotsinorder[$slotnumber]->page;
    }

    /**
     * Get the slot id of a given slot slot.
     * @param int $slotnumber the index of the slot in question.
     * @return int the page number of the page that slot is on.
     */
    public function get_slot_id_for_slot($slotnumber) {
        return $this->slotsinorder[$slotnumber]->id;
    }

    /**
     * Get the question type in a given slot.
     * @param int $slotnumber the index of the slot in question.
     * @return string the question type (e.g. multichoice).
     */
    public function get_question_type_for_slot($slotnumber) {
        return $this->questions[$this->slotsinorder[$slotnumber]->questionid]->qtype;
    }

    /**
     * Whether it would be possible, given the question types, etc. for the
     * question in the given slot to require that the previous question had been
     * answered before this one is displayed.
     * @param int $slotnumber the index of the slot in question.
     * @return bool can this question require the previous one.
     */
    public function can_question_depend_on_previous_slot($slotnumber) {
        return $slotnumber > 1 && $this->can_finish_during_the_attempt($slotnumber - 1);
    }

    /**
     * Whether it would be possible, given the question types, etc. for the
     * question in the given slot to require that the previous question had been
     * answered before this one is displayed.
     * @param int $slotnumber the index of the slot in question.
     * @return bool can this question require the previous one.
     */
    public function is_question_dependent_on_previous_slot($slotnumber) {
        return $this->slotsinorder[$slotnumber]->requireprevious;
    }

    /**
     * Is a particular question in this attempt a real question, or something like a description.
     * @param int $slotnumber the index of the slot in question.
     * @return bool whether that question is a real question.
     */
    public function is_real_question($slotnumber) {
        return $this->get_question_in_slot($slotnumber)->length != 0;
    }

    /**
     * Get the course module id of the prova.
     * @return int the course_modules.id for the prova.
     */
    public function get_cmid() {
        return $this->provaobj->get_cmid();
    }

    /**
     * Quizzes can only be repaginated if they have not been attempted, the
     * questions are not shuffled, and there are two or more questions.
     * @return bool whether this prova can be repaginated.
     */
    public function can_be_repaginated() {
        return $this->can_be_edited() && $this->get_question_count() >= 2;
    }

    /**
     * Get prova slots.
     * @return \stdClass[] the slots in this prova.
     */
    public function get_slots() {
        return array_column($this->slotsinorder, null, 'id');
    }

    /**
     * Is this slot the first one on its page?
     * @param int $slotnumber the index of the slot in question.
     * @return bool whether this slot the first one on its page.
     */
    public function is_first_slot_on_page($slotnumber) {
        if ($slotnumber == 1) {
            return true;
        }
        return $this->slotsinorder[$slotnumber]->page != $this->slotsinorder[$slotnumber - 1]->page;
    }

    /**
     * Is this slot the last one on its page?
     * @param int $slotnumber the index of the slot in question.
     * @return bool whether this slot the last one on its page.
     */
    public function is_last_slot_on_page($slotnumber) {
        if (!isset($this->slotsinorder[$slotnumber + 1])) {
            return true;
        }
        return $this->slotsinorder[$slotnumber]->page != $this->slotsinorder[$slotnumber + 1]->page;
    }

    /**
     * Is this slot the last one in its section?
     * @param int $slotnumber the index of the slot in question.
     * @return bool whether this slot the last one on its section.
     */
    public function is_last_slot_in_section($slotnumber) {
        return $slotnumber == $this->slotsinorder[$slotnumber]->section->lastslot;
    }

    /**
     * Is this slot the last one in the prova?
     * @param int $slotnumber the index of the slot in question.
     * @return bool whether this slot the last one in the prova.
     */
    public function is_last_slot_in_prova($slotnumber) {
        end($this->slotsinorder);
        return $slotnumber == key($this->slotsinorder);
    }

    /**
     * Is this the first section in the prova?
     * @param \stdClass $section the prova_sections row.
     * @return bool whether this is first section in the prova.
     */
    public function is_first_section($section) {
        return $section->firstslot == 1;
    }

    /**
     * Is this the last section in the prova?
     * @param \stdClass $section the prova_sections row.
     * @return bool whether this is first section in the prova.
     */
    public function is_last_section($section) {
        return $section->id == end($this->sections)->id;
    }


    /**
     * Get the final slot in the prova.
     * @return \stdClass the prova_slots for for the final slot in the prova.
     */
    public function get_last_slot() {
        return end($this->slotsinorder);
    }

    /**
     * Get a slot by it's id. Throws an exception if it is missing.
     * @param int $slotid the slot id.
     * @return \stdClass the requested prova_slots row.
     * @throws \coding_exception
     */
    public function get_slot_by_id($slotid) {
        foreach ($this->slotsinorder as $slot) {
            if ($slot->id == $slotid) {
                return $slot;
            }
        }

        throw new \coding_exception('The \'slotid\' could not be found.');
    }

    /**
     * Get a slot by it's slot number. Throws an exception if it is missing.
     *
     * @param int $slotnumber The slot number
     * @return \stdClass
     * @throws \coding_exception
     */
    public function get_slot_by_number($slotnumber) {
        if (!array_key_exists($slotnumber, $this->slotsinorder)) {
            throw new \coding_exception('The \'slotnumber\' could not be found.');
        }
        return $this->slotsinorder[$slotnumber];
    }

    /**
     * Check whether adding a section heading is possible
     * @param int $pagenumber the number of the page.
     * @return boolean
     */
    public function can_add_section_heading($pagenumber) {
        // There is a default section heading on this page,
        // do not show adding new section heading in the Add menu.
        if ($pagenumber == 1) {
            return false;
        }
        // Get an array of firstslots.
        $firstslots = array();
        foreach ($this->sections as $section) {
            $firstslots[] = $section->firstslot;
        }
        foreach ($this->slotsinorder as $slot) {
            if ($slot->page == $pagenumber) {
                if (in_array($slot->slot, $firstslots)) {
                    return false;
                }
            }
        }
        // Do not show the adding section heading on the last add menu.
        if ($pagenumber == 0) {
            return false;
        }
        return true;
    }

    /**
     * Get all the slots in a section of the prova.
     * @param int $sectionid the section id.
     * @return int[] slot numbers.
     */
    public function get_slots_in_section($sectionid) {
        $slots = array();
        foreach ($this->slotsinorder as $slot) {
            if ($slot->section->id == $sectionid) {
                $slots[] = $slot->slot;
            }
        }
        return $slots;
    }

    /**
     * Get all the sections of the prova.
     * @return \stdClass[] the sections in this prova.
     */
    public function get_sections() {
        return $this->sections;
    }

    /**
     * Get a particular section by id.
     * @return \stdClass the section.
     */
    public function get_section_by_id($sectionid) {
        return $this->sections[$sectionid];
    }

    /**
     * Get the number of questions in the prova.
     * @return int the number of questions in the prova.
     */
    public function get_section_count() {
        return count($this->sections);
    }

    public function populate_structure($provaobj) {
        global $DB;

        $prova = $provaobj->get_data();
        $slots = $DB->get_records_sql("
                SELECT slot.id AS slotid, slot.slot, slot.questionid, slot.page, slot.maxmark, seox.*, q.* 
                  FROM {seox_prova_slots} slot
                  LEFT JOIN {question} q ON q.id = slot.questionid
                  LEFT JOIN {seox_questions} seox ON seox.id_question = slot.questionid
                 WHERE slot.provaid = ?
              ORDER BY slot.slot", array($prova->id));

        $this->slotsinorder = array();
        foreach ($slots as $slotdata) {
            $this->questions[$slotdata->questionid] = $slotdata;

            $slot = new \stdClass();
            $slot->id = $slotdata->slotid;
            $slot->slot = $slotdata->slot;
            $slot->provaid = $prova->id;
            $slot->page = $slotdata->page;
            $slot->questionid = $slotdata->questionid;
            $slot->maxmark = $slotdata->maxmark;

            $this->slotsinorder[$slot->slot] = $slot;
        }

        // Get prova sections in ascending order of the firstslot.
        $this->populate_question_numbers();
    }

    /**
     * Number the questions.
     */
    protected function populate_question_numbers() {
        $number = 1;
        foreach ($this->slotsinorder as $slot) {
            if ($this->questions[$slot->questionid]->length == 0) {
                $slot->displayednumber = get_string('infoshort', 'prova');
            } else {
                $slot->displayednumber = $number;
                $number += 1;
            }
        }
    }

    /**
     * Recalcula a dificuldade de uma prova
     */
    public function recalculate_difficulty() {
        if ($this->can_be_edited()) {
            if ($newdif = difficulty_calculator($this->get_provaid(), $this->questions, $this->get_provadifficulty())) {
                $this->provaobj->get_data()->dificuldade = $newdif['result'];
            }
        }
    }

    /** --- General public functions */

    /** 
     * @return \int id da prova 
     */
    public function get_provadifficulty() {
        return $this->provaobj->get_data()->dificuldade;
    }

    /**
     * Retorna os parametros para criação/edicao/remoção de qeustões
     */
    public function get_course_params($returnurl=null) {
    	if (!$this->course_params) {
    		global $DB, $CFG;
    		include($CFG->dirroot . '/mod/prova/managelib.php');
    		if (!$category = $DB->get_record('question_categories', array('name' => 'top', 'contextid' => \context_course::instance(SEOX_COURSE_ID)->id))) {
	            print_error('error_getquestioncategory', 'provacomp_addquestions');
	        }
	        $params = array(
	            'returnurl' => $returnurl,
	            'courseid'  => SEOX_COURSE_ID,
	            'category'  => $category->id,
	        );
            $this->course_params = $params;
    	} elseif ($returnurl) {
    		$this->course_params['returnurl'] = $returnurl;
    	}
    	return $this->course_params;
    }

    /**
     * @return \int quantidade de paginas
     */
    public function get_pages_count() {
        return count($this->slotsinorder);
    }

    /** Overwritten */
    public function is_only_one_slot_in_section($section=null) {
    	return count($this->slotsinorder) === 1;
    }

    public function is_only_slot_in_section($section=null) {
        return $this->is_only_one_slot_in_section($section);
    }

    /** Overwritten */
    public function can_be_edited() {
    	return $this->provaobj->is_editable_prova();
    }

    /**
     * This prova can only be edited if they have not been attempted.
     * Throw an exception if this is not the case.
     */
    public function check_can_be_edited() {
        if (!$this->can_be_edited()) {
            print_error('err_noteditableprova', 'prova');
        }
    }

    /** Overwritten */
    public function get_courseid() {
    	global $CFG;
    	include($CFG->dirroot . '/mod/prova/managelib.php');
    	return SEOX_COURSE_ID;
    }

    /**
     * Get the prova object.
     * @return \stdClass the prova settings row from the database.
     */
    public function get_prova() {
        return $this->provaobj->get_prova();
    }

    /** 
     * @return \int id da prova 
     */
    public function get_provaid() {
        return $this->provaobj->get_data()->id;
    }

    /** Overwritten */
    public function get_slots_order() {
        $slots = array();
        foreach ($this->slotsinorder as $slot) {
            $slots[] = $slot->slot;
        }
        return $slots;
    }

    public function get_questions() {
        return $this->questions;
    }

    /** 
     * Verifica se slot existe
     *
     * @return \bool true caso exista, false caso nao
     */
    public function slot_exists($slot) {
        return array_key_exists($slot, $this->slotsinorder);
    }

    /** 
     * Verifica se ha slots na prova
     *
     * @return \bool true caso haja, false caso nao
     */
    public function is_not_empty() {
        return count($this->slotsinorder) > 0;
    }

    /** 
     * Overwritten 
     * Representa o número de casas decimais depois da vírgula
     */
    public function get_decimal_places_for_question_marks() {
        return 2;
    }

    /** 
     * Overwritten 
     * Representa o número de casas decimais da grade total
     */
    public function get_decimal_places_for_grades() {
        return 2;
    }

    public function question_can_be_edited($slotnumber) {
        $question = $this->get_question_in_slot($slotnumber);
        return $question->vezes_usada <= 0;
    }

    /** Overwritten */
    public function formatted_prova_grade() {
    	$grade = $this->provaobj->get_data()->grade;
        if (is_null($grade)) {
	        return get_string('notyetgraded', 'prova');
	    }
	    return format_float($grade, $this->get_decimal_places_for_question_marks());
    }

    /** Overwritten */
    public function get_questions_per_page() {
    	return 5;
    }

    /** Overwritten */
    public function can_finish_during_the_attempt($slotnumber=null) {
        return false;
    }

    /** Overwritten */
    public function formatted_question_grade($slotnumber) {
        return prova_format_question_grade($this->slotsinorder[$slotnumber]->maxmark);
    }

    /** Overwritten */
    public function move_slot($idmove, $idmoveafter, $page) {
        global $DB;

        $this->check_can_be_edited();

        $movingslot = $this->get_slot_by_id($idmove);
        if (empty($movingslot)) {
            throw new \moodle_exception('Bad slot ID ' . $idmove);
        }
        $movingslotnumber = (int) $movingslot->slot;

        // Empty target slot means move slot to first.
        if (empty($idmoveafter)) {
            $moveafterslotnumber = 0;
        } else {
            $moveafterslotnumber = (int) $this->get_slot_by_id($idmoveafter)->slot;
        }

        // If the action came in as moving a slot to itself, normalise this to
        // moving the slot to after the previous slot.
        if ($moveafterslotnumber == $movingslotnumber) {
            $moveafterslotnumber = $moveafterslotnumber - 1;
        }

        $followingslotnumber = $moveafterslotnumber + 1;
        // Prevent checking against non-existance slot when already at the last slot.
        if ($followingslotnumber == $movingslotnumber && !$this->is_last_slot_in_prova($followingslotnumber)) {
            $followingslotnumber += 1;
        }

        // Check the target page number is OK.
        if ($page == 0) {
            $page = 1;
        }
        if (($moveafterslotnumber > 0 && $page < $this->get_page_number_for_slot($moveafterslotnumber)) ||
                $page < 1) {
            throw new \coding_exception('The target page number is too small.');
        } else if (!$this->is_last_slot_in_prova($moveafterslotnumber) &&
                $page > $this->get_page_number_for_slot($followingslotnumber)) {
            throw new \coding_exception('The target page number is too large.');
        }

        // Work out how things are being moved.
        $slotreorder = array();
        if ($moveafterslotnumber > $movingslotnumber) {
            // Moving down.
            $slotreorder[$movingslotnumber] = $moveafterslotnumber;
            for ($i = $movingslotnumber; $i < $moveafterslotnumber; $i++) {
                $slotreorder[$i + 1] = $i;
            }

            $headingmoveafter = $movingslotnumber;
            if ($this->is_last_slot_in_prova($moveafterslotnumber) ||
                    $page == $this->get_page_number_for_slot($moveafterslotnumber + 1)) {
                // We are moving to the start of a section, so that heading needs
                // to be included in the ones that move up.
                $headingmovebefore = $moveafterslotnumber + 1;
            } else {
                $headingmovebefore = $moveafterslotnumber;
            }
            $headingmovedirection = -1;

        } else if ($moveafterslotnumber < $movingslotnumber - 1) {
            // Moving up.
            $slotreorder[$movingslotnumber] = $moveafterslotnumber + 1;
            for ($i = $moveafterslotnumber + 1; $i < $movingslotnumber; $i++) {
                $slotreorder[$i] = $i + 1;
            }

            if ($page == $this->get_page_number_for_slot($moveafterslotnumber + 1)) {
                // Moving to the start of a section, don't move that section.
                $headingmoveafter = $moveafterslotnumber + 1;
            } else {
                // Moving tot the end of the previous section, so move the heading down too.
                $headingmoveafter = $moveafterslotnumber;
            }
            $headingmovebefore = $movingslotnumber + 1;
            $headingmovedirection = 1;
        } else {
            // Staying in the same place, but possibly changing page/section.
            if ($page > $movingslot->page) {
                $headingmoveafter = $movingslotnumber;
                $headingmovebefore = $movingslotnumber + 2;
                $headingmovedirection = -1;
            } else if ($page < $movingslot->page) {
                $headingmoveafter = $movingslotnumber - 1;
                $headingmovebefore = $movingslotnumber + 1;
                $headingmovedirection = 1;
            } else {
                return; // Nothing to do.
            }
        }

        if ($this->is_only_slot_in_section($movingslotnumber)) {
            throw new \coding_exception('You cannot remove the last slot in a section.');
        }

        $trans = $DB->start_delegated_transaction();

        // Slot has moved record new order.
        if ($slotreorder) {
            update_field_with_unique_index('seox_prova_slots', 'slot', $slotreorder,
                    array('provaid' => $this->get_provaid()));
        }

        // Page has changed. Record it.
        if ($movingslot->page != $page) {
            $DB->set_field('seox_prova_slots', 'page', $page, array('provaid' => $this->get_provaid(), 'id' => $movingslot->id));
        }

        // Update section fist slots.
        // prova_update_section_firstslots($this->get_provaid(), $headingmovedirection, $headingmoveafter, $headingmovebefore);

        // If any pages are now empty, remove them.
        $emptypages = $DB->get_fieldset_sql("
                SELECT DISTINCT page - 1
                  FROM {seox_prova_slots} slot
                 WHERE provaid = ?
                   AND page > 1
                   AND NOT EXISTS (SELECT 1 FROM {seox_prova_slots} WHERE provaid = ? AND page = slot.page - 1)
              ORDER BY page - 1 DESC
                ", array($this->get_provaid(), $this->get_provaid()));

        foreach ($emptypages as $page) {
            $DB->execute("
                    UPDATE {seox_prova_slots}
                       SET page = page - 1
                     WHERE provaid = ?
                       AND page > ?
                    ", array($this->get_provaid(), $page));
        }

        $trans->allow_commit();
    }

    /** Overwritten */
    public function update_page_break($slotid, $type) {
        global $DB;

        $this->check_can_be_edited();

        $provaslots = $DB->get_records('seox_prova_slots', array('provaid' => $this->get_provaid()), 'slot');
        $repaginate = new \provacomp_addprovas\repaginate($this->get_provaid(), $provaslots);
        $repaginate->repaginate_slots($provaslots[$slotid]->slot, $type);
        $slots = $this->refresh_page_numbers_and_update_db();

        return $slots;
    }

    /** Overwritten */
    public function refresh_page_numbers_and_update_db() {
        global $DB;
        $this->check_can_be_edited();

        $slots = $this->refresh_page_numbers();

        // Record new page order.
        foreach ($slots as $slot) {
            $DB->set_field('seox_prova_slots', 'page', $slot->page,
                    array('id' => $slot->id));
        }

        return $slots;
    }

    /** Overwritten */
    public function refresh_page_numbers($slots = array()) {
        global $DB;
        // Get slots ordered by page then slot.
        if (!count($slots)) {
            $slots = $DB->get_records('seox_prova_slots', array('provaid' => $this->get_provaid()), 'slot, page');
        }

        // Loop slots. Start Page number at 1 and increment as required.
        $pagenumbers = array('new' => 0, 'old' => 0);

        foreach ($slots as $slot) {
            if ($slot->page !== $pagenumbers['old']) {
                $pagenumbers['old'] = $slot->page;
                ++$pagenumbers['new'];
            }

            if ($pagenumbers['new'] == $slot->page) {
                continue;
            }
            $slot->page = $pagenumbers['new'];
        }

        return $slots;
    }

    /** Overwritten */
    public function remove_slot($slotnumber) {
        global $DB;

        $this->check_can_be_edited();

        if ($this->is_only_slot_in_section($slotnumber) && $this->get_section_count() > 1) {
            throw new \coding_exception('You cannot remove the last slot in a section.');
        }

        $slot = $DB->get_record('seox_prova_slots', array('provaid' => $this->get_provaid(), 'slot' => $slotnumber));
        if (!$slot) {
            return;
        }
        $maxslot = $DB->get_field_sql('SELECT MAX(slot) FROM {seox_prova_slots} WHERE provaid = ?', array($this->get_provaid()));

        $trans = $DB->start_delegated_transaction();
        // $DB->delete_records('prova_slot_tags', array('slotid' => $slot->id));
        $DB->delete_records('seox_prova_slots', array('id' => $slot->id));
        for ($i = $slot->slot + 1; $i <= $maxslot; $i++) {
            $DB->set_field('seox_prova_slots', 'slot', $i - 1,
                    array('provaid' => $this->get_provaid(), 'slot' => $i));
            $this->slotsinorder[$i]->slot = $i - 1;
            $this->slotsinorder[$i - 1] = $this->slotsinorder[$i];
            unset($this->slotsinorder[$i]);
        }

        $qtype = $DB->get_field('question', 'qtype', array('id' => $slot->questionid));
        $vezes_usada = $DB->get_field('seox_questions', 'vezes_usada', array('id_question' => $slot->questionid));
        if ($vezes_usada > 0) {
            $DB->execute('UPDATE {seox_questions} SET vezes_usada = vezes_usada - 1 WHERE id_question = ?', [$slot->questionid]);
        }

        if ($qtype === 'random') {
            // This function automatically checks if the question is in use, and won't delete if it is.
            question_delete_question($slot->questionid);
        }

        $this->populate_question_numbers();
        unset($this->questions[$slot->questionid]);

        $this->refresh_page_numbers_and_update_db();

        $trans->allow_commit();
    }

    /**
     * Retorna a categoria das questões do curso seox
     */
    public static function get_cquestion_category_and_context_id($courseid) {
        global $DB, $CFG;
        require_once($CFG->dirroot . '/mod/prova/managelib.php');
        $contextid = \context_course::instance(SEOX_COURSE_ID)->id;
        if (!$category = $DB->get_record('question_categories', array('name' => 'top', 'contextid' => $contextid))) {
            print_error('error_getquestioncategory', 'provacomp_addquestions');
        }
        return $category->id . ','. $contextid;
    }

    /** Overwritten */
    public function update_slot_maxmark($slot, $maxmark) {
        global $DB;

        if (abs($maxmark - $slot->maxmark) < 1e-7) {
            // Grade has not changed. Nothing to do.
            return false;
        }

        $trans = $DB->start_delegated_transaction();
        $slot->maxmark = $maxmark;
        $DB->update_record('seox_prova_slots', $slot);
        $trans->allow_commit();

        return true;
    }

    /**
     * Set whether the question in a particular slot requires the previous one.
     * @param int $slotid id of slot.
     * @param bool $requireprevious if true, set this question to require the previous one.
     */
    public function update_question_dependency($slotid, $requireprevious) {
        global $DB;
        $DB->set_field('prova_slots', 'requireprevious', $requireprevious, array('id' => $slotid));
    }

    /**
     * Change the heading for a section.
     * @param int $id the id of the section to change.
     * @param string $newheading the new heading for this section.
     */
    public function set_section_heading($id, $newheading) {
        return;
    }

    /**
     * Change the shuffle setting for a section.
     * @param int $id the id of the section to change.
     * @param bool $shuffle whether this section should be shuffled.
     */
    public function set_section_shuffle($id, $shuffle) {
        return;
    }

    /**
     * Remove the section heading with the given id
     * @param int $sectionid the section to remove.
     */
    public function remove_section_heading($sectionid) {
        return;
    }
}