<?php 
/**
 * Model de uma Prova.
 * Based on: /mod/quiz/attemptlib.php and
 * /
 */
defined('MOODLE_INTERNAL') || die();

require_once ($CFG->dirroot . '/mod/prova/src/attempt/lib.php');

/**
 * Classe utilizada para gerar exception facilmente
 */ 
class moodle_prova_exception extends moodle_exception {
    public function __construct($provaobj, $errorcode, $a = null, $link = '', $debuginfo = null) {
        if (!$link) {
            $link = $provaobj->view_url();
        }
        parent::__construct($errorcode, 'provacomp_addprovas', $link, $a, $debuginfo);
    }
}

/**
 * Classe utilizada para disponibilizar fácil acesso
 * aos dados de uma prova específica
 */
class ProvaObj {

    /** @var stdClass dados da tabela seox_prova. */
    protected $prova;

    /** @var array de slots da prova. */
    protected $slots = null;

    /** @var array de questões extraidas dos slots. */
    protected $quastions = null;

    /** @bool verifica se usuário corrente é um professor editor. */
    protected $iseditingteacher = null;

    /** @bool se questão pode ser editada. */
    protected $iseditableprova = null;

    protected $structure = null;

    public function __construct($prova) {
        $this->prova = (is_object($prova)) ? $prova : $this->get_prova_from_db($prova);
        $this->get_slots_from_db();
        $this->get_questions_from_db();
        $this->verfify_teacher_permission();
        $this->verify_editable_prova();
    }

    public function get_prova() {
        return $this->get_data();
    } 

    /**
     * Pega uma prova do db com base no id
	 *
	 * @param $provaid id da prova
     */
    protected function get_prova_from_db($provaid) {
    	global $DB;
    	if (!$prova = $DB->get_record('seox_prova', array('id' => $provaid))) {
			throw new moodle_prova_exception($this, 'err_provanotfound', $provaid);
    	}
    	return $prova;
    }

    /**
     * Pega slots do db com base no id
     */
    protected function get_slots_from_db() {
    	global $DB;
    	if (!$this->prova || !$slots = $DB->get_records('seox_prova_slots', array('provaid' => $this->prova->id))) {
			return false;
    	}
    	$this->slots = $slots;
    }

    /**
     * Pega questions do db com base no id
     */
    protected function get_questions_from_db() {

    	if (!$this->slots) {
			return false;
    	}

    	global $DB;

    	$questionsids = array();
    	foreach ($this->slots as $slot) {
    		$questionsids[] = $slot->questionid;
    	}
    	if (!empty($questionsids)) {
    		$sql = 'SELECT 
    				q.id AS qid,
					seox.id_question AS seoxid,
					q.name AS qname,
					area.name AS s_area,
					subarea.name AS s_subarea,
					assunto.name AS s_assunto,
					q.questiontext AS qtext,
					seox.dificuldade AS dificuldade,
					seox.vezes_usada AS vezes_usada
				FROM {question} AS q 
				INNER JOIN {seox_questions} AS seox ON seox.id_question=q.id
				INNER JOIN {seox_area} AS area ON seox.id_area=area.id
				INNER JOIN {seox_subarea} AS subarea ON seox.id_subarea=subarea.id
				INNER JOIN {seox_assunto} AS assunto ON seox.id_assunto=assunto.id
				WHERE q.id IN ('. implode(',', $questionsids).');';
    		if (!$questions = $DB->get_records_sql($sql)) {
				throw new moodle_prova_exception($this, 'err_questionnotfound', $this->prova->id);
    		}
    		$this->questions = $questions;
    	} else {
    		$this->questions = false;
    	}
    }

    /**
     * Utiliza a capability do curso raiz para verificar permissoes
     */
    protected function verfify_teacher_permission() {
    	global $CFG;
    	require_once($CFG->dirroot . '/mod/prova/managelib.php');
    	$this->iseditingteacher = has_capability('mod/prova:accesstoread', context_course::instance(SEOX_COURSE_ID));
	}

	/**
	 * Verifica se a prova é editavel
	 */
	protected function verify_editable_prova() {
		if (!$this->prova) {
			throw new moodle_prova_exception($this, 'err_provanotfound', $this->prova->id);
		}
		$this->iseditableprova = $this->prova->vezes_usada <= 0;
	}

	/**
	 * --- GETTERS
	 */
	public function get_data() {
		return $this->prova;
	}

    public function get_provaid() {
        return $this->prova->id;
    }

	public function get_questions($questionids = null) {
        if (is_null($questionids)) {
            $questionids = array_keys($this->questions);
        }
        $questions = array();
        foreach ($questionids as $id) {
            if (!array_key_exists($id, $this->questions)) {
                throw new moodle_exception('cannotstartmissingquestion', 'prova', $this->view_url());
            }
            $questions[$id] = $this->questions[$id];
            $this->ensure_question_loaded($id);
        }
        return $questions;
    }

	public function is_editing_teacher() {
		return $this->iseditingteacher;
	}

	public function is_editable_prova() {
		return $this->iseditableprova;
	}
	

	/**
	 * --- LOGIC FUNCTIONS
	 * 
	 * Funções lógicas e frequêntes envolvento provas
	 */
    public function view_url() {
    	return new moodle_url('/mod/prova/components/provas.php');
    }

    /**
     * Cria e retorna um objeto do tipo structure
     */ 
    public function get_structure() {
        if (!$this->structure) {
            $struc = new \provacomp_addprovas\structure();
            $this->structure = $struc->create_for_prova($this);
        }
        return $this->structure;
    }
}

/*********** 
    ATTEMPT QUIZ TO PREVIEW PART 
                        ***********/
class prova_attempt_preview extends prova_attempt {

    protected $provaobj = null;

    protected $accessmanager = null;

    /** @var question_usage_by_activity the question usage for the preview. */
    protected $quba = null;

    /**
     * Inicia o class question_usage
     */
    public function __construct($provaid) {
        global $CFG, $USER;

        $this->provaobj = new ProvaObj($provaid);
        $this->provaobj->get_structure();

        $this->quba = question_engine::make_questions_usage_by_activity('provacomp_addprovas', context_user::instance($USER->id));

        $cont = 1;
        foreach ($this->provaobj->get_structure()->get_questions() as $slot => $question) {
            $newslot = $this->quba->add_question(question_bank::load_question($question->id), $question->maxmark);
            if ($newslot != $cont++) {
                throw new coding_exception('Slot numbers have got confused.');
            }
        }
        $this->quba->set_preferred_behaviour('deferredfeedback');
        $this->quba->start_all_questions();
    }

    public function get_prova() {
        return $this->provaobj->get_data();
    } 

    /**
     * Initialise the JS etc. required all the questions on a page.
     * @param mixed $page a page number, or 'all'.
     */
    public function get_html_head_contributions($page = 'all', $showall = false) {
        if ($showall) {
            $page = 'all';
        }
        $result = '';
        foreach ($this->get_slots($page) as $slot) {
            $result .= $this->quba->render_question_head_html($slot);
        }
        $result .= question_engine::initialise_js();
        return $result;
    }

    /**
     * @param integer page number.
     * @return array the requested list of slot numbers.
     */
    public function get_slots($page = 'all') {
        if ($page === 'all') {
            return $this->provaobj->get_structure()->get_slots_order();
        }
        if ($this->provaobj->get_structure()->slot_exists($page)) {
            return [$page];
        }
        return [];
    }

    /**
     * @return string the URL of this prova's edit page. Needs to be POSTed to with a cmid parameter.
     */
    public function start_attempt_url($slot = null, $page = -1) {
        return new moodle_url('/mod/prova/components/addprovas/preview.php', array('provaid' => $this->get_prova()->id));
    }

    /**
     * @param int $provaid the id of a prova.
     * @param int $page optional page number to go to in the attempt.
     * @return string the URL of that attempt.
     */
    public static function preview_url($provaid, $page = 0) {
        global $CFG;
        $url = $CFG->wwwroot . '/mod/prova/components/addprovas/preview.php?provaid=' . $provaid;
        if ($page) {
            $url .= '&page=' . $page;
        }
        return $url;
    }

    /**
     * Retorna a página que irá processar as ações na página de preview 
     */
    public static function processpreview_url($end=false) {
        global $CFG;
        $url = $CFG->wwwroot . '/mod/prova/components/addprovas/processpreview.php';
        if ($end) {
            $url .= '?page=-1';
        }
        return $url;
    }

    /**
     * Get the navigation panel object for this attempt.
     *
     * @param $panelclass The type of panel, prova_attempt_nav_panel or prova_review_nav_panel
     * @param $page the current page number.
     * @param $showall whether we are showing the whole prova on one page. (Used by review.php)
     * @return prova_nav_panel_base the requested object.
     */
    public function seox_get_navigation_panel(provacomp_addprovas_renderer $output, $panelclass, $page, $showall = false) {
        $panel = new $panelclass($this, prova_get_preview_options(), $page, $showall);

        $title = html_writer::tag('h5', get_string('provanavigation', 'provacomp_addprovas'), array('id' => 'mod_prova_navblock_title'));
        $content = $output->navigation_panel($panel, $title);
        return $content;
    }

    /**
     *
     * Generate the HTML that displayes the question in its current state, with
     * the appropriate display options.
     *
     * @param int $slot identifies the question in the attempt.
     * @param provacomp_addprovas_renderer $renderer the prova renderer.
     * @param moodle_url $thispageurl the URL of the page this question is being printed on.
     * @return string HTML for the question in its current state.
     */
    public function seox_render_question($slot, $renderer, $thispageurl = null) {
        return $this->seox_render_question_helper($slot, $thispageurl, $renderer, null);
    }

    /**
     * @param int $slot identifies the question in the attempt.
     * @param moodle_url $thispageurl the URL of the page this question is being printed on.
     * @param provacomp_addprovas_renderer $renderer the prova renderer.
     * @param int|null $seq the seq number of the past state to display.
     * @return string HTML fragment.
     */
    protected function seox_render_question_helper($slot, $thispageurl, $renderer, $seq) {
        $originalslot = $slot;
        $number = $this->get_question_number($originalslot);
        $displayoptions = prova_get_preview_options();

        // if ($slot != $originalslot) {
        //     $originalmaxmark = $this->get_question_attempt($slot)->get_max_mark();
        //     $this->get_question_attempt($slot)->set_max_mark($this->get_question_attempt($originalslot)->get_max_mark());
        // }

        // if ($this->can_question_be_redone_now($slot)) {
        //     $displayoptions->extrainfocontent = $renderer->redo_question_button($slot, $displayoptions->readonly);
        // }

        if ($seq === null) {
            $output = $this->quba->render_question($slot, $displayoptions, $slot);
        } else {
            $output = $this->quba->render_question_at_step($slot, $seq, $displayoptions, $number);
        }

        // if ($slot != $originalslot) {
        //     $this->get_question_attempt($slot)->set_max_mark($originalmaxmark);
        // }

        return $output;
    }

    /**
     * @return 1 \int representa o número da questão em uma página.
     */
    public function get_question_number($slot) {
        return 1;
    }

    /**
     * @param int $page the requested page number.
     * @return int a safe page number to use.
     */
    public function force_page_number_into_range($page) {
        return min(max($page, 1), $this->provaobj->get_structure()->get_pages_count());
    }

    /**
     * @param int $page the requested page number.
     * @return bool
     */
    public function is_last_page($page) {
        return $page == $this->provaobj->get_structure()->get_pages_count();
    }

    /**
     * If the section heading, if any, that should come just before this slot.
     * @param int $slot identifies a particular question in this attempt.
     * @return string the required heading, or null if there is not one here.
     */
    public function get_heading_before_slot($slot) {
        return null;
    }
}

/**
 * Represents the navigation panel, and builds a {@link block_contents} to allow
 * it to be output.
 */
abstract class addprova_nav_panel_base {

    /** @var prova_attempt */
    protected $attemptobj;
    /** @var question_display_options */
    protected $options;
    /** @var integer */
    protected $page;
    /** @var boolean */
    protected $showall;

    public function __construct($attemptobj, question_display_options $options, $page, $showall) {
        $this->attemptobj = $attemptobj;
        $this->options = $options;
        $this->page = $page;
        $this->showall = $showall;
    }

    /**
     * Get the buttons and section headings to go in the prova navigation block.
     * @return renderable[] the buttons, possibly interleaved with section headings.
     */
    public function get_question_buttons() {
        $buttons = array();
        foreach ($this->attemptobj->get_slots() as $slot) {
            // if ($heading = $this->attemptobj->get_heading_before_slot($slot)) {
            //     $buttons[] = new prova_nav_section_heading(format_string($heading));
            // }

            $qa = $this->attemptobj->get_question_attempt($slot);
            $showcorrectness = $this->options->correctness && $qa->has_marks();

            $button = new addprova_nav_question_button();
            $button->id          = 'quiznavbutton' . $slot;
            $button->number      = $slot;
            $button->stateclass  = $qa->get_state_class($showcorrectness);
            $button->navmethod   = 'free';
            if (!$showcorrectness && $button->stateclass == 'notanswered') {
                $button->stateclass = 'complete';
            }
            $button->statestring = $this->get_state_string($qa, $showcorrectness);
            $button->page        = $slot;
            $button->currentpage = $this->showall || $button->page == $this->page;
            $button->flagged     = $qa->is_flagged();
            $button->url         = $this->get_question_url($slot);
            if ($this->attemptobj->is_blocked_by_previous_question($slot)) {
                $button->url = null;
                $button->stateclass = 'blocked';
                $button->statestring = get_string('questiondependsonprevious', 'provacomp_addprovas');
            }
            $buttons[] = $button;
        }

        return $buttons;
    }

    protected function get_state_string(question_attempt $qa, $showcorrectness) {
        if ($qa->get_question()->length > 0) {
            return $qa->get_state_string($showcorrectness);
        }

        // Special case handling for 'information' items.
        if ($qa->get_state() == question_state::$todo) {
            return get_string('notyetviewed', 'prova');
        } else {
            return get_string('viewed', 'prova');
        }
    }

    public function render_before_button_bits(provacomp_addprovas_renderer $output) {
        return '';
    }

    abstract public function render_end_bits(provacomp_addprovas_renderer $output);

    protected function render_restart_preview_link($output) {
        return $output->restart_preview_button(new moodle_url($this->attemptobj->start_attempt_url(), array('forcenew' => true)));
    }

    protected abstract function get_question_url($slot);

    public function user_picture() {
        return null;
    }

    /**
     * Return 'allquestionsononepage' as CSS class name when $showall is set,
     * otherwise, return 'multipages' as CSS class name.
     * @return string, CSS class name
     */
    public function get_button_container_class() {
        // Quiz navigation is set on 'Show all questions on one page'.
        if ($this->showall) {
            return 'allquestionsononepage';
        }
        // Quiz navigation is set on 'Show one page at a time'.
        return 'multipages';
    }
}

/***/
class addprova_attempt_preview_nav_panel extends addprova_nav_panel_base {
    public function get_question_url($slot) {
        // if ($this->attemptobj->can_navigate_to($slot)) {
        if (true) {// mocado: true pois representa 'free' 
            return $this->attemptobj->preview_url($this->attemptobj->get_prova()->id, $slot);
        } else {
            return null;
        }
    }

    public function render_before_button_bits(provacomp_addprovas_renderer $output) {
        return html_writer::tag('div', get_string('navnojswarning', 'prova'),
                array('id' => 'quiznojswarning'));
    }

    public function render_end_bits(provacomp_addprovas_renderer $output) {
        return html_writer::link($this->attemptobj->processpreview_url(true),
                get_string('endtest', 'prova'), array('class' => 'endtestlink')) .
                $this->render_restart_preview_link($output);
    }
}

/***/
class addprova_nav_question_button implements renderable {
    /** @var string id="..." to add to the HTML for this button. */
    public $id;
    /** @var string number to display in this button. Either the question number of 'i'. */
    public $number;
    /** @var string class to add to the class="" attribute to represnt the question state. */
    public $stateclass;
    /** @var string Textual description of the question state, e.g. to use as a tool tip. */
    public $statestring;
    /** @var int the page number this question is on. */
    public $page;
    /** @var bool true if this question is on the current page. */
    public $currentpage;
    /** @var bool true if this question has been flagged. */
    public $flagged;
    /** @var moodle_url the link this button goes to, or null if there should not be a link. */
    public $url;
}


?>