<?php
/**
 * Library code used by prova cron.
 *
 * @package   mod_quiz
 */

defined('MOODLE_INTERNAL') || die();

require_once($CFG->dirroot . '/mod/prova/locallib.php');


/**
 * This class holds all the code for automatically updating all attempts that have
 * gone over their time limit.
 */
class mod_prova_overdue_attempt_updater {

    /**
     * Do the processing required.
     * @param int $timenow the time to consider as 'now' during the processing.
     * @param int $processto only process attempt with timecheckstate longer ago than this.
     * @return array with two elements, the number of attempt considered, and how many different quizzes that was.
     */
    public function update_overdue_attempts($timenow, $processto) {
        global $DB;

        $attemptstoprocess = $this->get_list_of_overdue_attempts($processto);

        $course = null;
        $prova = null;
        $cm = null;

        $count = 0;
        $provacount = 0;
        foreach ($attemptstoprocess as $attempt) {
            try {

                // If we have moved on to a different prova, fetch the new data.
                if (!$prova || $attempt->prova != $prova->id) {
                    $prova = $DB->get_record('prova', array('id' => $attempt->prova), '*', MUST_EXIST);
                    $cm = get_coursemodule_from_instance('prova', $attempt->prova);
                    $provacount += 1;
                }

                // If we have moved on to a different course, fetch the new data.
                if (!$course || $course->id != $prova->course) {
                    $course = $DB->get_record('course', array('id' => $prova->course), '*', MUST_EXIST);
                }

                // Make a specialised version of the prova settings, with the relevant overrides.
                $provaforuser = clone($prova);
                $provaforuser->timeclose = $attempt->usertimeclose;
                $provaforuser->timelimit = $attempt->usertimelimit;

                // Trigger any transitions that are required.
                $provaobj = prova::create($provaforuser->id);
                $attemptobj = new prova_attempt($attempt, $provaobj, $cm, $course);
                $attemptobj->handle_if_time_expired($timenow, false, true);
                $count += 1;

            } catch (moodle_exception $e) {
                // If an error occurs while processing one attempt, don't let that kill cron.
                mtrace("Error while processing attempt {$attempt->id} at {$attempt->prova} prova:");
                mtrace($e->getMessage());
                mtrace($e->getTraceAsString());
                // Close down any currently open transactions, otherwise one error
                // will stop following DB changes from being committed.
                $DB->force_transaction_rollback();
            }
        }

        $attemptstoprocess->close();
        return array($count, $provacount);
    }

    /**
     * @return moodle_recordset of quiz_attempts that need to be processed because time has
     *     passed. The array is sorted by courseid then quizid.
     */
    public function get_list_of_overdue_attempts($processto) {
        global $DB;

        // SQL to compute timeclose and timelimit for each attempt:
        $provaausersql = prova_get_attempt_usertime_sql(
                "iprovaa.state IN ('inprogress') AND iprovaa.timecheckstate <= :iprocessto");

        // This query should have all the quiz_attempts columns.
        return $DB->get_recordset_sql("
         SELECT provaa.*,
                provaauser.usertimeclose,
                provaauser.usertimelimit

           FROM {prova_attempts} provaa
           JOIN {prova} prova ON prova.id = provaa.prova
           JOIN ( $provaausersql ) provaauser ON provaauser.id = provaa.id

          WHERE provaa.state IN ('inprogress', 'overdue')
            AND provaa.timecheckstate <= :processto
       ORDER BY prova.course, provaa.prova",

                array('processto' => $processto, 'iprocessto' => $processto));
    }

}

// $teste = new mod_prova_overdue_attempt_updater();
// $teste->update_overdue_attempts(time(), time());
