<?php
/**
 * Biblioteca local de funções utilizadas pelo modulo prova
 *
 * @package /mod/prova
 */
defined('MOODLE_INTERNAL') || die();

require_once($CFG->dirroot . '/mod/prova/lib.php');
require_once($CFG->dirroot . '/mod/prova/renderer.php');
require_once($CFG->dirroot . '/mod/prova/src/attempt/lib.php');
require_once($CFG->dirroot . '/mod/prova/src/attempt/accessmanager.php');
require_once($CFG->dirroot . '/mod/prova/src/attempt/accessmanager_form.php');
// TODO: integrar com plguin
require_once($CFG->dirroot . '/mod/prova/components/addprovas/lib.php');

require_once($CFG->libdir . '/filelib.php');
require_once($CFG->libdir . '/questionlib.php');


/****************************************************
                Gerneral Functions
 ****************************************************/
/* Função que converte uma data em formato unix timestamp
*  para o formato DD-MM-YYYY H:M:s */
function format_date($date) {

    $date_param = new DateTime();
    $date_param->setTimestamp($date);
    $date_format = $date_param->format('d-m-Y H:i:s');

    return $date_format;
}


/****************************************************
                Addprovas Functions
 ****************************************************/
/**
 * Creates a textual representation of a question for display.
 *
 * @param object $question A question object from the database questions table
 * @param bool $showicon If true, show the question's icon with the question. False by default.
 * @param bool $showquestiontext If true (default), show question text after question name.
 *       If false, show only question name.
 * @return string
 */
function prova_question_tostring($question, $showicon = false, $showquestiontext = true) {
    $result = '';

    $name = shorten_text(format_string($question->name), 200);
    if ($showicon) {
        $name .= print_question_icon($question) . ' ' . $name;
    }
    $result .= html_writer::span($name, 'questionname');

    if ($showquestiontext) {
        $questiontext = question_utils::to_plain_text($question->questiontext,
                $question->questiontextformat, array('noclean' => true, 'para' => false));
        $questiontext = shorten_text($questiontext, 200);
        if ($questiontext) {
            $result .= ' ' . html_writer::span(s($questiontext), 'questiontext');
        }
    }

    return $result;
}

/****************************************************
                Prova Attempt Functions
 ****************************************************/
/**
 * @var int We show the countdown timer if there is less than this amount of time left before the
 * the prova close date. (1 hour)
 */
define('PROVA_SHOW_TIME_BEFORE_DEADLINE', '3600');

/**
 * @var int If there are fewer than this many seconds left when the student submits
 * a page of the prova, then do not take them to the next page of the prova. Instead
 * close the prova immediately.
 */
define('PROVA_MIN_TIME_TO_CONTINUE', '2');

/**
 * @var int We show no image when user selects No image from dropdown menu in prova settings.
 */
define('PROVA_SHOWIMAGE_NONE', 0);

/**
 * @var int We show small image when user selects small image from dropdown menu in prova settings.
 */
define('PROVA_SHOWIMAGE_SMALL', 1);

/**
 * @var int We show Large image when user selects Large image from dropdown menu in prova settings.
 */
define('PROVA_SHOWIMAGE_LARGE', 2);


/**
 * Function from /mod/quiz/locallib.php
 * Original name: quiz_create_attempt_handling_errors
 */
function prova_create_attempt_handling_errors($attemptid, $cmid = null) {
    try {
        $attempobj = prova_attempt::create($attemptid);
    } catch (moodle_exception $e) {
        if (!empty($cmid)) {
            list($course, $cm) = get_course_and_cm_from_cmid($cmid, 'prova');
            $continuelink = new moodle_url('/mod/prova/view.php', array('id' => $cmid));
            $context = context_module::instance($cm->id);
            if (has_capability('mod/prova:preview', $context)) {
                throw new moodle_exception('attempterrorcontentchange', 'prova', $continuelink);
            } else {
                throw new moodle_exception('attempterrorcontentchangeforuser', 'prova', $continuelink);
            }
        } else {
            throw new moodle_exception('attempterrorinvalid', 'prova');
        }
    }
    if (!empty($cmid) && $attempobj->get_cmid() != $cmid) {
        throw new moodle_exception('invalidcoursemodule');
    } else {
        return $attempobj;
    }
}

/**
 * Function from /mod/quiz/locallib.php
 * Original name: quiz_create_attempt
 */
function prova_create_attempt(prova $provaobj, $attemptnumber, $lastattempt, $timenow, $ispreview = false, $userid = null) {
    global $USER;

    if ($userid === null) {
        $userid = $USER->id;
    }

    $provarootobj = $provaobj->get_provaobj();
    $prova = $provaobj->get_prova();
    if ($provarootobj->sumgrades < 0.000005 && $provarootobj->grade > 0.000005) {
        throw new moodle_exception('cannotstartgradesmismatch', 'prova',
                new moodle_url('/mod/prova/view.php', array('q' => $prova->id)), array('grade' => modprova_format_grade($provarootobj, $provarootobj->grade)));
    }

    $attempt = new stdClass();
    $attempt->prova = $prova->id;
    $attempt->userid = $userid;
    $attempt->preview = 0;
    $attempt->layout = '';
    $attempt->attempt = $attemptnumber;
    $attempt->timestart = $timenow;
    $attempt->timefinish = 0;
    $attempt->timemodified = $timenow;
    $attempt->timemodifiedoffline = 0;
    $attempt->state = prova_attempt::IN_PROGRESS;
    $attempt->currentpage = 0;
    $attempt->sumgrades = null;

    // If this is a preview, mark it as such.
    if ($ispreview) {
        $attempt->preview = 1;
    }

    $timeclose = $provaobj->get_access_manager($timenow)->get_end_time($attempt);
    if ($timeclose === false || $ispreview) {
        $attempt->timecheckstate = null;
    } else {
        $attempt->timecheckstate = $timeclose;
    }

    return $attempt;
}

/**
 * Function from /mod/quiz/locallib.php
 * Original name: quiz_start_new_attempt
 */
function prova_start_new_attempt($provaobj, $quba, $attempt, $attemptnumber, $timenow,
                                $questionids = array(), $forcedvariantsbyslot = array()) {

    // Usages for this user's previous prova attempts.
    $qubaids = new \mod_prova\question\qubaids_for_users_attempts(
            $provaobj->get_provaid(), $attempt->userid);

    // Fully load all the questions in this prova.
    $provaobj->preload_questions();
    $provaobj->load_questions();

    // First load all the non-random questions.
    $randomfound = false;
    $slot = 0;
    $questions = array();
    $maxmark = array();
    $page = array();
    foreach ($provaobj->get_questions() as $questiondata) {
        $slot += 1;
        $maxmark[$slot] = $questiondata->maxmark;
        $page[$slot] = $questiondata->page;
 
        if (!$provaobj->get_prova()->shuffleanswers) {
            $questiondata->options->shuffleanswers = false;
        }
        $questions[$slot] = question_bank::make_question($questiondata);
    }

    // Finally add them all to the usage.
    ksort($questions);
    foreach ($questions as $slot => $question) {
        $newslot = $quba->add_question($question, $maxmark[$slot]);
        if ($newslot != $slot) {
            throw new coding_exception('Slot numbers have got confused.');
        }
    }

    // Start all the questions.
    $variantstrategy = new core_question\engine\variants\least_used_strategy($quba, $qubaids);

    if (!empty($forcedvariantsbyslot)) {
        $forcedvariantsbyseed = question_variant_forced_choices_selection_strategy::prepare_forced_choices_array(
            $forcedvariantsbyslot, $quba);
        $variantstrategy = new question_variant_forced_choices_selection_strategy(
            $forcedvariantsbyseed, $variantstrategy);
    }

    $quba->start_all_questions($variantstrategy, $timenow);

    return $attempt;
}

/**
 * Function from /mod/quiz/locallib.php
 * Original name: quiz_validate_new_attempt
 */
function prova_validate_new_attempt(prova $provaobj, prova_access_manager $accessmanager, $forcenew, $page, $redirect) {
    global $DB, $USER;
    $timenow = time();

    if ($provaobj->is_preview_user() && $forcenew) {
        $accessmanager->current_attempt_finished();
    }

    // Check capabilities.
    if (!$provaobj->is_preview_user()) {
        $provaobj->require_capability('mod/prova:attempt');
    }

    // Check to see if a new preview was requested.
    if ($provaobj->is_preview_user() && $forcenew) {
        // To force the creation of a new preview, we mark the current attempt (if any)
        // as abandoned. It will then automatically be deleted below.
        $DB->set_field('prova_attempts', 'state', prova_attempt::ABANDONED,
                array('prova' => $provaobj->get_provaid(), 'userid' => $USER->id));
    }

    // Look for an existing attempt.
    $attempts = prova_get_user_attempts($provaobj->get_provaid(), $USER->id, 'all', true);
    $lastattempt = end($attempts);

    $attemptnumber = null;
    // If an in-progress attempt exists, check password then redirect to it.
    if ($lastattempt && ($lastattempt->state == prova_attempt::IN_PROGRESS || $lastattempt->state == prova_attempt::OVERDUE)) {
        $currentattemptid = $lastattempt->id;
        $messages = $accessmanager->prevent_access();

        // If the attempt is now overdue, deal with that.
        $provaobj->create_attempt_object($lastattempt)->handle_if_time_expired($timenow, true);

        // And, if the attempt is now no longer in progress, redirect to the appropriate place.
        if ($lastattempt->state == prova_attempt::ABANDONED || $lastattempt->state == prova_attempt::FINISHED) {
            if ($redirect) {
                redirect($provaobj->review_url($lastattempt->id));
            } else {
                throw new moodle_prova_exception($provaobj, 'attemptalreadyclosed');
            }
        }

        // If the page number was not explicitly in the URL, go to the current page.
        if ($page == -1) {
            $page = $lastattempt->currentpage;
        }

    } else {
        while ($lastattempt && $lastattempt->preview) {
            $lastattempt = array_pop($attempts);
        }

        // Get number for the next or unfinished attempt.
        if ($lastattempt) {
            $attemptnumber = $lastattempt->attempt + 1;
        } else {
            $lastattempt = false;
            $attemptnumber = 1;
        }
        $currentattemptid = null;

        $messages = $accessmanager->prevent_access() + $accessmanager->prevent_new_attempt(count($attempts), $lastattempt);

        if ($page == -1) {
            $page = 0;
        }
    }
    return array($currentattemptid, $attemptnumber, $lastattempt, $messages, $page);
}

/**
 * Function from /mod/quiz/locallib.php
 * Original name: prova_attempt_save_started
 */
function prova_attempt_save_started($provaobj, $quba, $attempt) {
    global $DB;
    // Save the attempt in the database.
    question_engine::save_questions_usage_by_activity($quba);
    $attempt->uniqueid = $quba->get_id();
    $attempt->id = $DB->insert_record('prova_attempts', $attempt);

    // Params used by the events below.
    $params = array(
        'objectid' => $attempt->id,
        'relateduserid' => $attempt->userid,
        'courseid' => $provaobj->get_courseid(),
        'context' => $provaobj->get_context()
    );
    // Decide which event we are using.
    // if ($attempt->preview) {
    //     $params['other'] = array(
    //         'provaid' => $provaobj->get_provaid()
    //     );
    //     $event = \mod_prova\event\attempt_preview_started::create($params);
    // } else {
        $event = \mod_prova\event\attempt_started::create($params);
    // }

    // Trigger the event.
    $event->add_record_snapshot('prova', $provaobj->get_prova());
    $event->add_record_snapshot('prova_attempts', $attempt);
    $event->trigger();

    return $attempt;
}

/**
 * Function from /mod/quiz/locallib.php
 * Original name: quiz_prepare_and_start_new_attempt
 */
function prova_prepare_and_start_new_attempt(prova $provaobj, $attemptnumber, $lastattempt,
        $offlineattempt = false, $forcedrandomquestions = [], $forcedvariants = []) {
    global $DB, $USER;

    // Delete any previous preview attempts belonging to this user.
    prova_delete_previews($provaobj->get_prova(), $USER->id);

    $quba = question_engine::make_questions_usage_by_activity('mod_prova', $provaobj->get_context());
    $quba->set_preferred_behaviour($provaobj->get_prova()->preferredbehaviour);

    // Create the new attempt and initialize the question sessions
    $timenow = time(); // Update time now, in case the server is running really slowly.
    $attempt = prova_create_attempt($provaobj, $attemptnumber, $lastattempt, $timenow, $provaobj->is_preview_user());
    $attempt = prova_start_new_attempt($provaobj, $quba, $attempt, $attemptnumber, $timenow, $forcedrandomquestions, $forcedvariants);

    $transaction = $DB->start_delegated_transaction();

    // Init the timemodifiedoffline for offline attempts.
    if ($offlineattempt) {
        $attempt->timemodifiedoffline = $attempt->timemodified;
    }
    $attempt = prova_attempt_save_started($provaobj, $quba, $attempt);

    $transaction->allow_commit();

    return $attempt;
}

/**
 * Function from /mod/quiz/locallib.php
 * Original name: quiz_delete_previews
 */
function prova_delete_previews($prova, $userid = null) {
    global $DB;
    $conditions = array('prova' => $prova->id, 'preview' => 1);
    if (!empty($userid)) {
        $conditions['userid'] = $userid;
    }
    $previewattempts = $DB->get_records('prova_attempts', $conditions);
    foreach ($previewattempts as $attempt) {
        prova_delete_attempt($attempt, $prova);
    }
}

/**
 * Function from /mod/quiz/locallib.php
 * Original name: quiz_delete_attempt
 */
function prova_delete_attempt($attempt, $prova) {
    global $DB;
    if (is_numeric($attempt)) {
        if (!$attempt = $DB->get_record('prova_attempts', array('id' => $attempt))) {
            return;
        }
    }

    if ($attempt->prova != $prova->id) {
        debugging("Trying to delete attempt $attempt->id which belongs to prova $attempt->prova " .
                "but was passed prova $prova->id.");
        return;
    }

    if (!isset($prova->cmid)) {
        $cm = get_coursemodule_from_instance('prova', $prova->id, $prova->course);
        $prova->cmid = $cm->id;
    }

    question_engine::delete_questions_usage_by_activity($attempt->uniqueid);
    $DB->delete_records('prova_attempts', array('id' => $attempt->id));

    // TODO: preparar log de provas
    // Log the deletion of the attempt if not a preview.
    // if (!$attempt->preview) {
    //     $params = array(
    //         'objectid' => $attempt->id,
    //         'relateduserid' => $attempt->userid,
    //         'context' => context_module::instance($prova->cmid),
    //         'other' => array(
    //             'provaid' => $prova->id
    //         )
    //     );
    //     $event = \mod_prova\event\attempt_deleted::create($params);
    //     $event->add_record_snapshot('prova_attempts', $attempt);
    //     $event->trigger();
    // }

    // TODO: verificar necessida de grades
    // Search prova_attempts for other instances by this user.
    // If none, then delete record for this prova, this user from prova_grades
    // else recalculate best grade.
    // $userid = $attempt->userid;
    // if (!$DB->record_exists('prova_attempts', array('userid' => $userid, 'prova' => $prova->id))) {
    //     $DB->delete_records('prova_grades', array('userid' => $userid, 'prova' => $prova->id));
    // } else {
    //     prova_save_best_grade($prova, $userid);
    // }

    // prova_update_grades($prova, $userid);
}

/**
 * Function from /mod/quiz/locallib.php
 * Original name: quiz_save_best_grade
 */
function prova_save_best_grade($prova, $userid = null, $attempts = array()) {
    global $DB, $OUTPUT, $USER;

    if (empty($userid)) {
        $userid = $USER->id;
    }

    if (empty($attempts)) {
        // Get all the attempts made by the user.
        $attempts = prova_get_user_attempts($prova->id, $userid);
    }

    // Calculate the best grade.
    $bestgrade = prova_calculate_best_grade($prova, $attempts);

    // Save the best grade in the database.
    if (is_null($bestgrade)) {
        $DB->delete_records('prova_grades', array('prova' => $prova->id, 'userid' => $userid));

    } else if ($grade = $DB->get_record('prova_grades',
            array('prova' => $prova->id, 'userid' => $userid))) {
        
        $grade->grade = $bestgrade;
        $grade->timemodified = time();
        $DB->update_record('prova_grades', $grade);

    } else {
        $grade = new stdClass();
        $grade->prova = $prova->id;
        $grade->userid = $userid;
        $grade->grade = $bestgrade;
        $grade->timemodified = time();
        $DB->insert_record('prova_grades', $grade);
    }
}

/**
 * Function from /mod/quiz/locallib.php
 * Original name: quiz_calculate_best_grade
 */
function prova_calculate_best_grade($prova, $attempts) {

    switch ($prova->grademethod) {

        case PROVA_ATTEMPTFIRST:
            $firstattempt = reset($attempts);
            return $firstattempt->sumgrades;

        case PROVA_ATTEMPTLAST:
            $lastattempt = end($attempts);
            return $lastattempt->sumgrades;

        case PROVA_GRADEAVERAGE:
            $sum = 0;
            $count = 0;
            foreach ($attempts as $attempt) {
                if (!is_null($attempt->sumgrades)) {
                    $sum += $attempt->sumgrades;
                    $count++;
                }
            }
            if ($count == 0) {
                return null;
            }
            return $sum / $count;

        case PROVA_GRADEHIGHEST:
        default:
            $max = null;
            foreach ($attempts as $attempt) {
                if ($attempt->sumgrades > $max) {
                    $max = $attempt->sumgrades;
                }
            }
            return $max;
    }
}

/**
 * Get the information about the standard prova JavaScript module.
 * @return array a standard jsmodule structure.
 */
function prova_get_js_module() {
    global $PAGE;

    return array(
        'name' => 'mod_prova',
        'fullpath' => '/mod/prova/src/attempt/module.js',
        'requires' => array('base', 'dom', 'event-delegate', 'event-key',
                'core_question_engine', 'moodle-core-formchangechecker'),
        'strings' => array(
            array('cancel', 'moodle'),
            array('flagged', 'question'),
            array('functiondisabledbysecuremode', 'prova'),
            array('startattempt', 'prova'),
            array('timesup', 'prova'),
            array('changesmadereallygoaway', 'moodle'),
        ),
    );
}

/**
 * An extension of question_display_options that includes the extra options used
 * by the prova.
 *
 * @copyright  2010 The Open University
 */
class mod_prova_display_options extends question_display_options {
    /**#@+
     * @var integer bits used to indicate various times in relation to a
     * prova attempt.
     */
    const DURING =            mod_prova_root_display_options::DURING;
    const IMMEDIATELY_AFTER = mod_prova_root_display_options::IMMEDIATELY_AFTER;
    const LATER_WHILE_OPEN =  mod_prova_root_display_options::LATER_WHILE_OPEN;
    const AFTER_CLOSE =       mod_prova_root_display_options::AFTER_CLOSE;
    /**#@-*/

    /**
     * @var boolean if this is false, then the student is not allowed to review
     * anything about the attempt.
     */
    public $attempt = true;

    /**
     * @var boolean if this is false, then the student is not allowed to review
     * anything about the attempt.
     */
    public $overallfeedback = self::VISIBLE;

    /**
     * Set up the various options from the prova settings, and a time constant.
     * @param object $prova the prova settings.
     * @param int $one of the {@link DURING}, {@link IMMEDIATELY_AFTER},
     * {@link LATER_WHILE_OPEN} or {@link AFTER_CLOSE} constants.
     * @return mod_prova_display_options set up appropriately.
     */
    public static function make_from_prova($prova, $when) {
        $options = new self();

        $options->attempt = self::VISIBLE;//self::extract($prova->reviewattempt, $when, true, false);
        $options->correctness = self::extract($prova->reviewcorrectness, $when);
        $options->marks = self::extract($prova->reviewmarks, $when, self::MARK_AND_MAX, self::MAX_ONLY);
        $options->feedback = self::extract($prova->reviewspecificfeedback, $when);
        $options->generalfeedback = self::extract($prova->reviewgeneralfeedback, $when);
        $options->rightanswer = self::extract($prova->reviewrightanswer, $when);
        $options->overallfeedback = self::extract($prova->reviewoverallfeedback, $when);

        $options->numpartscorrect = $options->feedback;
        $options->manualcomment = $options->feedback;

        if ($prova->questiondecimalpoints != -1) {
            $options->markdp = $prova->questiondecimalpoints;
        } else {
            $options->markdp = $prova->decimalpoints;
        }

        return $options;
    }

    protected static function extract($bitmask, $bit,
            $whenset = self::VISIBLE, $whennotset = self::HIDDEN) {
        if ($bitmask & $bit) {
            return $whenset;
        } else {
            return $whennotset;
        }
    }
}

/**
 * Work out what state this prova attempt is in - in the sense used by
 * prova_get_review_options, not in the sense of $attempt->state.
 * @param object $prova the prova settings
 * @param object $attempt the prova_attempt database row.
 * @return int one of the mod_prova_display_options::DURING,
 *      IMMEDIATELY_AFTER, LATER_WHILE_OPEN or AFTER_CLOSE constants.
 */
function prova_attempt_state($prova, $attempt) {
    if ($prova->timeclose && time() >= $prova->timeclose) {
        return mod_prova_display_options::AFTER_CLOSE;
    } else {
        return mod_prova_display_options::DURING;
    } 
}

/**
 * The the appropraite mod_prova_display_options object for this attempt at this
 * prova right now.
 *
 * @param object $prova the prova instance.
 * @param object $attempt the attempt in question.
 * @param $context the prova context.
 *
 * @return mod_prova_display_options
 */
function prova_get_review_options($prova, $attempt, $context, $printter, $correcting) {
    global $USER;
    $options = mod_prova_display_options::make_from_prova($prova, prova_attempt_state($prova, $attempt));

    $options->readonly = true;
    $options->flags = question_display_options::HIDDEN;
    if (!empty($attempt->id)) {
        $options->questionreviewlink = new moodle_url('/mod/prova/screens/attempt/reviewquestion.php', array('attempt' => $attempt->id));
    }

    // Show a link to the comment box only for closed attempts.
    if ($correcting && !$printter && !empty($attempt->id) && $attempt->state == prova_attempt::FINISHED && !$attempt->preview &&
            !is_null($context) && has_capability('mod/prova:grade', $context)) {
        $options->manualcomment = question_display_options::VISIBLE;
        $options->manualcommentlink = new moodle_url('/mod/prova/screens/report/comment.php',
                array('attempt' => $attempt->id));
    }


    if (!is_null($context) && !$attempt->preview) {
        $options->attempt = question_display_options::VISIBLE;

        if ($prova->corrected && $attempt->userid == $USER->id) {
            $options->correctness = question_display_options::VISIBLE;
            $options->marks = question_display_options::MARK_AND_MAX;
            $options->feedback = question_display_options::VISIBLE;
            $options->numpartscorrect = question_display_options::VISIBLE;
            $options->manualcomment = question_display_options::VISIBLE;
            $options->rightanswer = question_display_options::VISIBLE;
        }
    }

    if (!$printter && !is_null($context) && !$attempt->preview &&
            has_capability('mod/prova:viewreports', $context) &&
            has_capability('moodle/grade:viewhidden', $context)) {
        // People who can see reports and hidden grades should be shown everything,
        // except during preview when teachers want to see what students see.
        // $options->attempt = question_display_options::VISIBLE;
        $options->correctness = question_display_options::VISIBLE;
        $options->marks = question_display_options::MARK_AND_MAX;
        $options->feedback = question_display_options::VISIBLE;
        $options->numpartscorrect = question_display_options::VISIBLE;
        $options->manualcomment = question_display_options::VISIBLE;
        $options->generalfeedback = question_display_options::VISIBLE;
        $options->rightanswer = question_display_options::VISIBLE;
        $options->overallfeedback = question_display_options::VISIBLE;
        $options->history = question_display_options::VISIBLE;

    }

    return $options;
}

/**
 * Send the notification message when a prova attempt becomes overdue.
 *
 * @param prova_attempt $attemptobj all the data about the prova attempt.
 */
function prova_send_overdue_message($attemptobj) {
    global $CFG, $DB;

    $submitter = $DB->get_record('user', array('id' => $attemptobj->get_userid()), '*', MUST_EXIST);

    if (!$attemptobj->has_capability('mod/prova:emailwarnoverdue', $submitter->id, false)) {
        return; // Message not required.
    }

    if (!$attemptobj->has_response_to_at_least_one_graded_question()) {
        return; // Message not required.
    }

    // Prepare lots of useful information that admins might want to include in
    // the email message.
    $provaname = format_string($attemptobj->get_prova_name());

    $deadlines = array();
    if ($attemptobj->get_prova()->get_prova()->timelimit) {
        $deadlines[] = $attemptobj->get_attempt()->timestart + $attemptobj->get_prova()->get_prova()->timelimit;
    }
    if ($attemptobj->get_prova()->get_prova()->timeclose) {
        $deadlines[] = $attemptobj->get_prova()->get_prova()->timeclose;
    }
    $duedate = min($deadlines);
    $graceend = $duedate + $attemptobj->get_prova()->get_prova()->graceperiod;

    $a = new stdClass();
    // Course info.
    $a->courseid           = $attemptobj->get_course()->id;
    $a->coursename         = format_string($attemptobj->get_course()->fullname);
    $a->courseshortname    = format_string($attemptobj->get_course()->shortname);
    // Quiz info.
    $a->provaname           = $provaname;
    $a->provaurl            = $attemptobj->view_url();
    $a->provalink           = '<a href="' . $a->provaurl . '">' . $provaname . '</a>';
    // Attempt info.
    $a->attemptduedate     = userdate($duedate);
    $a->attemptgraceend    = userdate($graceend);
    $a->attemptsummaryurl  = $attemptobj->summary_url()->out(false);
    $a->attemptsummarylink = '<a href="' . $a->attemptsummaryurl . '">' . $provaname . ' review</a>';
    // Student's info.
    $a->studentidnumber    = $submitter->idnumber;
    $a->studentname        = fullname($submitter);
    $a->studentusername    = $submitter->username;

    // Prepare the message.
    $eventdata = new \core\message\message();
    $eventdata->courseid          = $a->courseid;
    $eventdata->component         = 'mod_prova';
    $eventdata->name              = 'attempt_overdue';
    $eventdata->notification      = 1;

    $eventdata->userfrom          = core_user::get_noreply_user();
    $eventdata->userto            = $submitter;
    $eventdata->subject           = get_string('emailoverduesubject', 'prova', $a);
    $eventdata->fullmessage       = get_string('emailoverduebody', 'prova', $a);
    $eventdata->fullmessageformat = FORMAT_PLAIN;
    $eventdata->fullmessagehtml   = '';

    $eventdata->smallmessage      = get_string('emailoverduesmall', 'prova', $a);
    $eventdata->contexturl        = $a->provaurl;
    $eventdata->contexturlname    = $a->provaname;
    $eventdata->customdata        = [
        'cmid' => $attemptobj->get_cmid(),
        'instance' => $attemptobj->get_provaid(),
        'attemptid' => $attemptobj->get_attemptid(),
    ];

    // Send the message.
    return message_send($eventdata);
}

/**
 * Convert the raw grade stored in $attempt into a grade out of the maximum
 * grade for this prova.
 *
 * @param float $rawgrade the unadjusted grade, fof example $attempt->sumgrades
 * @param object $prova the prova object. Only the fields grade, sumgrades and decimalpoints are used.
 * @param bool|string $format whether to format the results for display
 *      or 'question' to format a question grade (different number of decimal places.
 * @return float|string the rescaled grade, or null/the lang string 'notyetgraded'
 *      if the $grade is null.
 */
function prova_rescale_grade($rawgrade, $prova, $format = true) {
    if (is_null($rawgrade)) {
        $grade = null;
    } else if ($prova->sumgrades >= 0.000005) {
        $grade = $rawgrade * $prova->sumgrades / $prova->sumgrades;
    } else {
        $grade = 0;
    }
    if ($format === 'question') {
        $grade = modprova_format_question_grade($prova, $grade);
    } else if ($format) {
        $grade = modprova_format_grade($prova, $grade);
    }
    return $grade;
}

/**
 * Get the human-readable name for a prova attempt state.
 * @param string $state one of the state constants like {@link prova_attempt::IN_PROGRESS}.
 * @return string The lang string to describe that state.
 */
function prova_attempt_state_name($state) {
    switch ($state) {
        case prova_attempt::IN_PROGRESS:
            return get_string('stateinprogress', 'prova');
        case prova_attempt::OVERDUE:
            return get_string('stateoverdue', 'prova');
        case prova_attempt::FINISHED:
            return get_string('statefinished', 'prova');
        case prova_attempt::ABANDONED:
            return get_string('stateabandoned', 'prova');
        default:
            throw new coding_exception('Estado da prova desconhecido.');
    }
}

/**
 * Returns an unfinished attempt (if there is one) for the given
 * user on the given prova. This function does not return preview attempts.
 *
 * @param int $provaid the id of the prova.
 * @param int $userid the id of the user.
 *
 * @return mixed the unfinished attempt if there is one, false if not.
 */
function prova_get_user_attempt_unfinished($provaid, $userid) {
    $attempts = prova_get_user_attempts($provaid, $userid, 'unfinished', true);
    if ($attempts) {
        return array_shift($attempts);
    } else {
        return false;
    }
}

/**
 * Combines the review options from a number of different prova attempts.
 * Returns an array of two ojects, so the suggested way of calling this
 * funciton is:
 * list($someoptions, $alloptions) = prova_get_combined_reviewoptions(...)
 *
 * @param object $prova the prova instance.
 * @param array $attempts an array of attempt objects.
 *
 * @return array of two options objects, one showing which options are true for
 *          at least one of the attempts, the other showing which options are true
 *          for all attempts.
 */
function prova_get_combined_reviewoptions($prova, $attempts) {
    $fields = array('feedback', 'generalfeedback', 'rightanswer', 'overallfeedback');
    $someoptions = new stdClass();
    $alloptions = new stdClass();
    foreach ($fields as $field) {
        $someoptions->$field = false;
        $alloptions->$field = true;
    }
    $someoptions->marks = question_display_options::HIDDEN;
    $alloptions->marks = question_display_options::MARK_AND_MAX;

    // This shouldn't happen, but we need to prevent reveal information.
    if (empty($attempts)) {
        return array($someoptions, $someoptions);
    }

    foreach ($attempts as $attempt) {
        $attemptoptions = mod_prova_display_options::make_from_prova($prova,
                prova_attempt_state($prova, $attempt));
        foreach ($fields as $field) {
            $someoptions->$field = $someoptions->$field || $attemptoptions->$field;
            $alloptions->$field = $alloptions->$field && $attemptoptions->$field;
        }
        $someoptions->marks = max($someoptions->marks, $attemptoptions->marks);
        $alloptions->marks = min($alloptions->marks, $attemptoptions->marks);
    }
    return array($someoptions, $alloptions);
}

// Functions to do with prova grades ////////////////////////////////////////////

/**
 * @return array int => lang string the options for calculating the prova grade
 *      from the individual attempt grades.
 */
function prova_get_grading_options() {
    return array(
        PROVA_GRADEHIGHEST => get_string('gradehighest', 'prova'),
        PROVA_GRADEAVERAGE => get_string('gradeaverage', 'prova'),
        PROVA_ATTEMPTFIRST => get_string('attemptfirst', 'prova'),
        PROVA_ATTEMPTLAST  => get_string('attemptlast', 'prova')
    );
}

/**
 * @param int $option one of the values PROVA_GRADEHIGHEST, PROVA_GRADEAVERAGE,
 *      PROVA_ATTEMPTFIRST or PROVA_ATTEMPTLAST.
 * @return the lang string for that option.
 */
function prova_get_grading_option_name($option) {
    $strings = prova_get_grading_options();
    return $strings[$option];
}

/**
 * Update the sumgrades field of the attempts at a quiz.
 *
 * @param object $prova a prova.
 */
function prova_update_all_attempt_sumgrades($prova) {
    global $DB;
    $dm = new question_engine_data_mapper();
    $timenow = time();

    $sql = "UPDATE {prova_attempts}
            SET
                timemodified = :timenow,
                sumgrades = (
                    {$dm->sum_usage_marks_subquery('uniqueid')}
                )
            WHERE prova = :provaid AND state = :finishedstate";
    $DB->execute($sql, array('timenow' => $timenow, 'provaid' => $prova->id,
            'finishedstate' => prova_attempt::FINISHED));
}

/**
 * Update the final grade at this prova for all students.
 *
 * This function is equivalent to calling prova_save_best_grade for all
 * users, but much more efficient.
 *
 * @param object $prova the prova settings.
 */
function prova_update_all_final_grades($prova) {
    global $DB;

    if (!$prova->sumgrades) {
        return;
    }

    $param = array('iprovaid' => $prova->id, 'istatefinished' => prova_attempt::FINISHED);
    $firstlastattemptjoin = "JOIN (
            SELECT
                iprova_a.userid,
                MIN(attempt) AS firstattempt,
                MAX(attempt) AS lastattempt

            FROM {prova_attempts} iprova_a

            WHERE
                iprova_a.state = :istatefinished AND
                iprova_a.preview = 0 AND
                iprova_a.prova = :iprovaid

            GROUP BY iprova_a.userid
        ) first_last_attempts ON first_last_attempts.userid = prova_a.userid";

    switch ($prova->grademethod) {
        case PROVA_ATTEMPTFIRST:
            // Because of the where clause, there will only be one row, but we
            // must still use an aggregate function.
            $select = 'MAX(prova_a.sumgrades)';
            $join = $firstlastattemptjoin;
            $where = 'prova_a.attempt = first_last_attempts.firstattempt AND';
            break;

        case PROVA_ATTEMPTLAST:
            // Because of the where clause, there will only be one row, but we
            // must still use an aggregate function.
            $select = 'MAX(prova_a.sumgrades)';
            $join = $firstlastattemptjoin;
            $where = 'prova_a.attempt = first_last_attempts.lastattempt AND';
            break;

        case PROVA_GRADEAVERAGE:
            $select = 'AVG(prova_a.sumgrades)';
            $join = '';
            $where = '';
            break;

        default:
        case PROVA_GRADEHIGHEST:
            $select = 'MAX(prova_a.sumgrades)';
            $join = '';
            $where = '';
            break;
    }

    if ($prova->sumgrades >= 0.000005) {
        $finalgrade = $select . ' * ' . ($prova->grade / $prova->sumgrades);
    } else {
        $finalgrade = '0';
    }
    $param['provaid'] = $prova->id;
    $param['provaid2'] = $prova->id;
    $param['provaid3'] = $prova->id;
    $param['provaid4'] = $prova->id;
    $param['statefinished'] = prova_attempt::FINISHED;
    $param['statefinished2'] = prova_attempt::FINISHED;
    $finalgradesubquery = "
            SELECT prova_a.userid, $finalgrade AS newgrade
            FROM {prova_attempts} prova_a
            $join
            WHERE
                $where
                prova_a.state = :statefinished AND
                prova_a.preview = 0 AND
                prova_a.prova = :provaid3
            GROUP BY prova_a.userid";

    $changedgrades = $DB->get_records_sql("
            SELECT users.userid, pg.id, pg.grade, newgrades.newgrade

            FROM (
                SELECT userid
                FROM {prova_grades} pg
                WHERE prova = :provaid
            UNION
                SELECT DISTINCT userid
                FROM {prova_attempts} prova_a2
                WHERE
                    prova_a2.state = :statefinished2 AND
                    prova_a2.preview = 0 AND
                    prova_a2.prova = :provaid2
            ) users

            LEFT JOIN {prova_grades} pg ON pg.userid = users.userid AND pg.prova = :provaid4

            LEFT JOIN (
                $finalgradesubquery
            ) newgrades ON newgrades.userid = users.userid

            WHERE
                ABS(newgrades.newgrade - pg.grade) > 0.000005 OR
                ((newgrades.newgrade IS NULL OR qg.grade IS NULL) AND NOT
                          (newgrades.newgrade IS NULL AND pg.grade IS NULL))",
                // The mess on the previous line is detecting where the value is
                // NULL in one column, and NOT NULL in the other, but SQL does
                // not have an XOR operator, and MS SQL server can't cope with
                // (newgrades.newgrade IS NULL) <> (qg.grade IS NULL).
            $param);

    $timenow = time();
    $todelete = array();
    foreach ($changedgrades as $changedgrade) {

        if (is_null($changedgrade->newgrade)) {
            $todelete[] = $changedgrade->userid;

        } else if (is_null($changedgrade->grade)) {
            $toinsert = new stdClass();
            $toinsert->prova = $prova->id;
            $toinsert->userid = $changedgrade->userid;
            $toinsert->timemodified = $timenow;
            $toinsert->grade = $changedgrade->newgrade;
            $DB->insert_record('prova_grades', $toinsert);

        } else {
            $toupdate = new stdClass();
            $toupdate->id = $changedgrade->id;
            $toupdate->grade = $changedgrade->newgrade;
            $toupdate->timemodified = $timenow;
            $DB->update_record('prova_grades', $toupdate);
        }
    }

    if (!empty($todelete)) {
        list($test, $params) = $DB->get_in_or_equal($todelete);
        $DB->delete_records_select('prova_grades', 'prova = ? AND userid ' . $test,
                array_merge(array($prova->id), $params));
    }
}

/** 
 * Retorna botão de corrgir caso professor tenha
 * permissao
 *
 * @param int $provaid, id da prova
 * @param prova $prova, objeto da classe prova
 * @return Str, html do botão de corrigir 
 */
function render_report_button($provaid, $prova=null) {
    global $USER, $OUTPUT;

    $provaobj = ($prova) ? $prova : prova::create($provaid, $USER->id);

    if($provaobj->has_capability('mod/prova:grade')) {
        $url = new moodle_url('/mod/prova/screens/report/report.php', array('id' => $provaobj->get_cmid(), 'mode' => 'responses'));
        $corrected = $provaobj->get_prova()->corrected;
        $correctness = ($corrected) ? 'corrected' : '';
        $title = ($corrected) ? get_string('viewtablecorrectbtn_corrected', 'prova') : get_string('viewtablecorrectbtn', 'prova');
        $button =  new single_button(
            $url, 
            $title,
            'get',
            array('class' => 'viewcorrectbtn')
        );

        $output = '<span class="viewcorrectbtn '.$correctness.'">';
        $output .= $OUTPUT->render($button);
        $output .= '</span>';
    
        return $output;
    } else {
        return '';
    }
}

/**
 * Atualiza a dificuldade de cada questão com
 * prova com base na última públicação de notas.
 */
function update_questions_difficulties($arraydiff) {
    global $DB;

    foreach ($arraydiff as $keydiff) {
        $questionobj = new stdClass();
        $questionobj->questionid = $keydiff['questionid'];
        $questionobj->dificuldade = $keydiff['value'];
        $questionobj->qa = $keydiff['qa'];
        $questionobj->qr = $keydiff['qr'];
        $questionobj->timemodified = time();

        //Insere uma linha para cada questõa na tabela seox_questions_difficulty
        $DB->insert_record('seox_questions_difficulty', $questionobj);

        //Atualização de dificuldade da questão na tabela seox_questions
        if($questionobj->dificuldade <= 0 || $questionobj->dificuldade < 0.34){

            $dificuldade = 0;
           
        }elseif($questionobj->dificuldade <= 0.34 || $questionobj->dificuldade < 0.67){

            $dificuldade = 1;

        }elseif($questionobj->dificuldade > 0.67){

            $dificuldade = 2;
        }

	   $DB->execute("UPDATE {seox_questions} SET dificuldade = $dificuldade WHERE id_question = ?", [$questionobj->questionid]);
        
    }

}

/**
 * Atualiza as nota dos alunos que responderam 
 * e finalizaram a prova de uma aplicação
 * a qual teve uma questão anulada.
 */
function update_sumgrades_question_canceled($attempt, $slot){

    global $DB;
    
    //Retorna a aplicação da questão anulada
    $prova_attempt_obj = $DB->get_record('prova_attempts', array('id' => $attempt));
    $prova = intval($prova_attempt_obj->prova);

    //Retorna o id da prova da questão anulada e o campo cancelled_questions
    $table_prova = $DB->get_record('prova', array('id' => $prova));

    //Retorna o id e a nota máxima da questão anulada
    $table_slots = $DB->get_record('seox_prova_slots', array('provaid' => $table_prova->provaid, 'slot' => $slot), 'id, maxmark, questionid');

    //Verifica se o campo cancelled_questions da tabela mdl_prova está vazio
    $cancelled_qtns = ($table_prova->cancelled_questions) ? unserialize($table_prova->cancelled_questions) : [];
    if (empty($cancelled_qtns)) {
        $id_anulados = array(); 
        $id_anulados[] = $table_slots->questionid; 
        $anulados_ids = serialize($id_anulados); 
        $status = true;
    } else {
        // Verificar se existe no array o id da questão
        if (in_array($table_slots->questionid, $cancelled_qtns)) {
            $status = false;
        } else {
            $cancelled_qtns[] = $table_slots->questionid;
            $anulados_ids = serialize($cancelled_qtns); 
            $status = true;
        }
    }

    if ($status) {
        $transaction = $DB->start_delegated_transaction();

        //Atualiza o campo cancelled_questions na tabela mdl_prova
        $DB->execute("UPDATE {prova} SET timemodified = ?, cancelled_questions = ? WHERE id = ?", [time(), $anulados_ids, $prova]);
        $uaids = $DB->get_records('prova_attempts', array('prova' => $prova), null, 'uniqueid, id');

        update_tables_to_cancell_question($table_slots->questionid, $table_slots->maxmark, $uaids);

        foreach ($uaids as $obj) {
            $attempt_obj = prova_attempt::create($obj->id);
            $quba = $attempt_obj->get_quba();

            $quba_slots = $quba->get_slots();
            $mark = 0;
            foreach ($quba_slots as $slot) {
                $qa = $quba->get_question_attempt($slot);
                $qid = $qa->get_question()->id;
                $inarray = in_array($qid, $cancelled_qtns);

                if (!$inarray && $qa->get_max_mark() > 0 && $qa->get_state() == question_state::$needsgrading) {
                    $mark = null;
                    break;
                }
                
                if ($inarray) {
                    $mark += $table_slots->maxmark;
                } else {
                    $mark += $qa->get_mark();
                }

            }
            
            if ($mark) {
                $DB->execute("UPDATE {prova_attempts} SET sumgrades = ? WHERE id = ?", [$mark, $obj->id]);
                $DB->execute("UPDATE {prova_grades} SET grade = ? WHERE prova = ? AND userid = ?", [
                    $mark, 
                    $attempt_obj->get_provaid(),
                    $attempt_obj->get_userid()
                ]);
            }
        }

        prova_save_best_grade($table_prova, $prova_attempt_obj->userid);

        $transaction->allow_commit();
    }

    return $status;
}

function update_sumgrades_question_canceled_at_attempt_process_finish($attempt, $prova, $quba) {
    global $DB;

    $cancelled_qtns = ($prova->cancelled_questions) ? unserialize($prova->cancelled_questions) : [];
    $cancelled_data = [];

    $table_slots = $DB->get_records('seox_prova_slots', array('provaid' => intval($prova->provaid)/* , 'questionid' => implode(', ', $cancelled_qtns) */), null, 'id, maxmark, questionid');
    
    foreach ($table_slots as $slot) {
        if (in_array($slot->questionid, $cancelled_qtns)) {
            update_tables_to_cancell_question($slot->questionid, $slot->maxmark, [$attempt->uniqueid => $attempt->uniqueid]);
            $cancelled_data[$slot->questionid] = $slot->maxmark;
        }
    }

    $quba_slots = $quba->get_slots();
    $mark = 0;
    foreach ($quba_slots as $slot) {
        $qa = $quba->get_question_attempt($slot);
        $qid = $qa->get_question()->id;
        $inarray = in_array($qid, $cancelled_qtns);

        if (!$inarray && $qa->get_max_mark() > 0 && $qa->get_state() == question_state::$needsgrading) {
            return null;
        }
        
        if ($inarray) {
            $mark += $cancelled_data[$qid];
        } else {
            $mark += $qa->get_mark();
        }
    }

    return $mark;
}

function update_tables_to_cancell_question($questionid, $maxmark, $uaids) {
    global $DB, $USER;

    $uaids = array_keys($uaids);
    $uaids = implode(',', $uaids);
    $qatts_ids = $DB->get_records_sql('SELECT * FROM {question_attempts} 
        WHERE questionid = ? AND questionusageid IN (' . $uaids . ')', [$questionid]);
    $qatts_ids = array_keys($qatts_ids);
    $qatts_ids_implode = implode(', ', $qatts_ids);

    $sequencenumbers = $DB->get_records_sql('SELECT questionattemptid, MAX(sequencenumber) AS sequencenumber FROM {question_attempt_steps} WHERE questionattemptid IN (' . $qatts_ids_implode . ') GROUP BY questionattemptid');

    // Adicionar steps para cada alunos que responder a questão
    $qatts_steps = [];
    foreach ($qatts_ids as $id) {
        $qatts_steps[] = array(
            'questionattemptid' => $id,
            'sequencenumber' => $sequencenumbers[$id]->sequencenumber + 1,
            'state' => 'mangrright', 
            'fraction' => 1,
            'timecreated' => time(),
            'userid' => $USER->id
        );
    }
    $DB->insert_records('question_attempt_steps', $qatts_steps);

    // Adicionar um metadata de comentário nos steps
    $qatts_ids = implode(', ', $qatts_ids);
    $newsteps_ids = $DB->get_records_sql('SELECT MAX(id), questionattemptid, MAX(sequencenumber) AS sequencenumber FROM {question_attempt_steps} WHERE questionattemptid IN (' . $qatts_ids_implode . ') GROUP BY questionattemptid');
    $newsteps_ids = array_keys($newsteps_ids);

    $comment = get_string('canceledquestion_comment', 'prova', number_format($maxmark, 2));
    $qatts_steps_metadata = [];
    foreach ($newsteps_ids as $id) {
        $qatts_steps_metadata[] = array(
            'attemptstepid' => $id,
            'name' => '-comment',
            'value' => $comment, 
        );
        $qatts_steps_metadata[] = array(
            'attemptstepid' => $id,
            'name' => '-maxmark',
            'value' => 1, 
        );
        $qatts_steps_metadata[] = array(
            'attemptstepid' => $id,
            'name' => '-mark',
            'value' => 1, 
        );
        $qatts_steps_metadata[] = array(
            'attemptstepid' => $id,
            'name' => '-commentformat',
            'value' => 1, 
        );
    }
    $DB->insert_records('question_attempt_step_data', $qatts_steps_metadata);
}

/**
 * Função responsável por finalizar prova e registrar em um comentário que
 * o usário colou
 */
function handle_provacontrol_action($attempt, $page) {
    global $CFG, $DB;

    if ($attempt->get_state() != prova_attempt::IN_PROGRESS) { return; }

    $time = time();

    $attempt->process_finish($time, false);

    // Inicia transicao personalizada
    $transaction = $DB->start_delegated_transaction();

    $quba = $attempt->get_quba();
    $qas = [];
    foreach ($quba->get_slots() as $slot) {
        $qa = $quba->get_question_attempt($slot);
        $qas[] = $qa->get_database_id();
    }
    $qas_implode = implode(', ', $qas);

    // Adicionar stepmetas com comments para os últimos steps
    $stepdata = $DB->get_records_sql('SELECT MAX(id) AS id, questionattemptid, MAX(sequencenumber) AS sequencenumber FROM {question_attempt_steps} WHERE questionattemptid IN (' . $qas_implode . ') GROUP BY questionattemptid');

    $comment = get_string('provacontrol_comment', 'prova', array('page' => $page, 'time' => format_date($time)));

    $qatts_steps_metadata = [];
    foreach ($stepdata as $id => $nil) {
        $qatts_steps_metadata[] = array(
            'attemptstepid' => $id,
            'name' => '-comment',
            'value' => $comment, 
        );
    }

    $DB->insert_records('question_attempt_step_data', $qatts_steps_metadata);
    
    $transaction->allow_commit();

    return $sequencenumbers;
}

function redirect_url($url, $message='', $delay=null, $messagetype = \core\output\notification::NOTIFY_INFO) {
    global $OUTPUT, $PAGE, $CFG;

    if (CLI_SCRIPT or AJAX_SCRIPT) {
        // This is wrong - developers should not use redirect in these scripts but it should not be very likely.
        throw new moodle_exception('redirecterrordetected', 'error');
    }

    if ($delay === null) {
        $delay = -1;
    }

    // Prevent debug errors - make sure context is properly initialised.
    if ($PAGE) {
        $PAGE->set_context(null);
        $PAGE->set_pagelayout('redirect');  // No header and footer needed.
        $PAGE->set_title(get_string('pageshouldredirect', 'moodle'));
    }

    if ($url instanceof moodle_url) {
        $url = $url->out(false);
    }

    $debugdisableredirect = false;
    do {
        if (defined('DEBUGGING_PRINTED')) {
            // Some debugging already printed, no need to look more.
            $debugdisableredirect = true;
            break;
        }

        if (core_useragent::is_msword()) {
            // Clicking a URL from MS Word sends a request to the server without cookies. If that
            // causes a redirect Word will open a browser pointing the new URL. If not, the URL that
            // was clicked is opened. Because the request from Word is without cookies, it almost
            // always results in a redirect to the login page, even if the user is logged in in their
            // browser. This is not what we want, so prevent the redirect for requests from Word.
            $debugdisableredirect = true;
            break;
        }

        if (empty($CFG->debugdisplay) or empty($CFG->debug)) {
            // No errors should be displayed.
            break;
        }

        if (!function_exists('error_get_last') or !$lasterror = error_get_last()) {
            break;
        }

        if (!($lasterror['type'] & $CFG->debug)) {
            // Last error not interesting.
            break;
        }

        // Watch out here, @hidden() errors are returned from error_get_last() too.
        if (headers_sent()) {
            // We already started printing something - that means errors likely printed.
            $debugdisableredirect = true;
            break;
        }

        if (ob_get_level() and ob_get_contents()) {
            // There is something waiting to be printed, hopefully it is the errors,
            // but it might be some error hidden by @ too - such as the timezone mess from setup.php.
            $debugdisableredirect = true;
            break;
        }
    } while (false);

    // Technically, HTTP/1.1 requires Location: header to contain the absolute path.
    // (In practice browsers accept relative paths - but still, might as well do it properly.)
    // This code turns relative into absolute.
    if (!preg_match('|^[a-z]+:|i', $url)) {
        // Get host name http://www.wherever.com.
        $hostpart = preg_replace('|^(.*?[^:/])/.*$|', '$1', $CFG->wwwroot);
        if (preg_match('|^/|', $url)) {
            // URLs beginning with / are relative to web server root so we just add them in.
            $url = $hostpart.$url;
        } else {
            // URLs not beginning with / are relative to path of current script, so add that on.
            $url = $hostpart.preg_replace('|\?.*$|', '', me()).'/../'.$url;
        }
        // Replace all ..s.
        while (true) {
            $newurl = preg_replace('|/(?!\.\.)[^/]*/\.\./|', '/', $url);
            if ($newurl == $url) {
                break;
            }
            $url = $newurl;
        }
    }

    // Sanitise url - we can not rely on moodle_url or our URL cleaning
    // because they do not support all valid external URLs.
    $url = preg_replace('/[\x00-\x1F\x7F]/', '', $url);
    $url = str_replace('"', '%22', $url);
    $encodedurl = preg_replace("/\&(?![a-zA-Z0-9#]{1,8};)/", "&amp;", $url);
    $encodedurl = preg_replace('/^.*href="([^"]*)".*$/', "\\1", clean_text('<a href="'.$encodedurl.'" />', FORMAT_HTML));
    $url = str_replace('&amp;', '&', $encodedurl);

    if (!empty($message)) {
        if (!$debugdisableredirect && !headers_sent()) {
            // A message has been provided, and the headers have not yet been sent.
            // Display the message as a notification on the subsequent page.
            \core\notification::add($message, $messagetype);
            $message = null;
            $delay = 0;
        } else {
            if ($delay === -1 || !is_numeric($delay)) {
                $delay = 3;
            }
            $message = clean_text($message);
        }
    } else {
        $message = get_string('pageshouldredirect');
        $delay = 0;
    }

    // Make sure the session is closed properly, this prevents problems in IIS
    // and also some potential PHP shutdown issues.
    \core\session\manager::write_close();

    if ($delay == 0 && !$debugdisableredirect && !headers_sent()) {
        // 302 might not work for POST requests, 303 is ignored by obsolete clients.
        @header($_SERVER['SERVER_PROTOCOL'] . ' 303 See Other');
        @header('Location: '.$url);
        echo bootstrap_renderer::plain_redirect_message($encodedurl);
        exit;
    }
    
    // Include a redirect message, even with a HTTP redirect, because that is recommended practice.
    if ($PAGE) {
        $debugdisableredirect = false;
        $CFG->docroot = false; // To prevent the link to moodle docs from being displayed on redirect page.
        echo $OUTPUT->redirect_message($encodedurl, $message, $delay, $debugdisableredirect, $messagetype);
        exit;
    } else {
        echo bootstrap_renderer::early_redirect_message($encodedurl, $message, $delay);
        exit;
    }
}

/**
 * A {@link qubaid_condition} for finding all the question usages belonging to
 * a particular prova.
 *
 * @copyright  2010 The Open University
 * @license    http://www.gnu.org/copyleft/gpl.html GNU GPL v3 or later
 */
class qubaids_for_prova extends qubaid_join {
    public function __construct($provaid, $includepreviews = true, $onlyfinished = false) {
        $where = 'provaa.prova = :provaaprova';
        $params = array('provaaprova' => $provaid);

        if (!$includepreviews) {
            $where .= ' AND preview = 0';
        }

        if ($onlyfinished) {
            $where .= ' AND state = :statefinished';
            $params['statefinished'] = prova_attempt::FINISHED;
        }

        parent::__construct('{prova_attempts} provaa', 'provaa.uniqueid', $where, $params);
    }
}

/**
 * Returns SQL to compute timeclose and timelimit for every attempt, taking into account user and group overrides.
 * The query used herein is very similar to the one in function prova_get_user_timeclose, so, in case you
 * would change either one of them, make sure to apply your changes to both.
 *
 * @param string $redundantwhereclauses extra where clauses to add to the subquery
 *      for performance. These can use the table alias iprovaa for the prova attempts table.
 * @return string SQL select with columns attempt.id, usertimeclose, usertimelimit.
 */
function prova_get_attempt_usertime_sql($redundantwhereclauses = '') {
    if ($redundantwhereclauses) {
        $redundantwhereclauses = 'WHERE ' . $redundantwhereclauses;
    }
    // The multiple qgo JOINS are necessary because we want timeclose/timelimit = 0 (unlimited) to supercede
    // any other group override
    $provaausersql = "
        SELECT iprovaa.id,
            iprova.timeclose AS usertimeclose,
            iprova.timelimit AS usertimelimit

        FROM {prova_attempts} iprovaa
        JOIN {prova} iprova ON iprova.id = iprovaa.prova
        LEFT JOIN {groups_members} gm ON gm.userid = iprovaa.userid
            $redundantwhereclauses
        GROUP BY iprovaa.id, iprova.id, iprova.timeclose, iprova.timelimit";

    return $provaausersql;
}
