<?php
/**
 * This page prints a review of a particular prova attempt
 *
 * It is used either by the student whose attempts this is, after the attempt,
 * or by a teacher reviewing another's attempt during or afterwards.
 *
 * @package   mod_prova
 */

require_once(__DIR__ . '/../config.php');
require_once($CFG->dirroot . '/mod/prova/locallib.php');

$attemptid = required_param('attempt', PARAM_INT);
$page      = optional_param('page', 0, PARAM_INT);
$showall   = optional_param('showall', null, PARAM_BOOL);
$cmid      = optional_param('cmid', null, PARAM_INT);
$printter   = optional_param('printter', false, PARAM_BOOL);

$url = new moodle_url('/mod/prova/screens/attempt/review.php', array('attempt'=>$attemptid));
if ($page !== 0) {
    $url->param('page', $page);
} else if ($showall) {
    $url->param('showall', $showall);
}
$PAGE->set_url($url);

$attemptobj = prova_create_attempt_handling_errors($attemptid, $cmid);
$page = $attemptobj->force_page_number_into_range($page);

// Now we can validate the params better, re-genrate the page URL.
if ($showall === null) {
    $showall = $page == 0 && $attemptobj->get_default_show_all('review');
}
$PAGE->set_url($attemptobj->review_url(null, $page, $showall));

// Check login.
require_login($attemptobj->get_course(), false, $attemptobj->get_cm());
$attemptobj->check_review_capability();

// Create an object to manage all the other (non-roles) access rules.
$accessmanager = $attemptobj->get_access_manager(time());
$accessmanager->setup_attempt_page($PAGE);

$options = $attemptobj->get_display_options(true, $printter);

// Check permissions - warning there is similar code in reviewquestion.php and
// prova_attempt::check_file_access. If you change on, change them all.
if ($attemptobj->is_own_attempt()) {
    if (!$attemptobj->is_finished()) {
        redirect($attemptobj->attempt_url(null, $page));
        
    } else if (!$options->attempt) {
        $accessmanager->back_to_view_page($PAGE->get_renderer('mod_prova'), $attemptobj->cannot_review_message());
       
    }

} else if (!$attemptobj->is_review_allowed()) {
    throw new moodle_provaattempt_exception($attemptobj->get_prova(), 'noreviewattempt');
}

// Load the questions and states needed by this page.
if ($showall) {
    $questionids = $attemptobj->get_slots();
} else {
    $questionids = $attemptobj->get_slots($page);
}

// Work out appropriate title and whether blocks should be shown.
if ($attemptobj->is_own_preview()) {
    $strreviewtitle = get_string('reviewofpreview', 'prova');
    // navigation_node::override_active_url($attemptobj->start_attempt_url());

} else {
    $strreviewtitle = get_string('reviewofattempt', 'prova', $attemptobj->get_attempt_number());
    if (empty($attemptobj->get_prova()->get_prova()->showblocks) && !$attemptobj->is_preview_user()) {
        $PAGE->blocks->show_only_fake_blocks();
    }
}

// Set up the page header.
$headtags = $attemptobj->get_html_head_contributions($page, $showall);
$PAGE->set_title($attemptobj->review_page_title($page, $showall));
$PAGE->set_heading($attemptobj->get_course()->fullname);

// Summary table start. ============================================================================

// Work out some time-related things.
$attempt = $attemptobj->get_attempt();
$prova = $attemptobj->get_prova()->get_prova();
$overtime = 0;

if ($attempt->state == prova_attempt::FINISHED) {
    if ($timetaken = ($attempt->timefinish - $attempt->timestart)) {
        if ($prova->timelimit && $timetaken > ($prova->timelimit + 60)) {
            $overtime = $timetaken - $prova->timelimit;
            $overtime = format_time($overtime);
        }
        $timetaken = format_time($timetaken);
    } else {
        $timetaken = "-";
    }
} else {
    $timetaken = get_string('unfinished', 'prova');
}

// Prepare summary informat about the whole attempt.
$summarydata = array();

$summarydata['coursetitle'] = array(
    'title'   => get_string('attemptcoursenamelabel', 'prova'),
    'content' => $attemptobj->review_print_course_title(),
);

$summarydata['maintitle'] = array(
    'title'   => get_string('attemptnamelabel', 'prova'),
    'content' => $attemptobj->review_print_title(),
);

$summarydata['usertitle'] = array(
    'title'   => get_string('usernamelabel', 'prova'),
    'content' => $attemptobj->user_print_title(),
);

if ($attemptobj->has_capability('mod/prova:viewreports')) {
    $attemptlist = $attemptobj->links_to_other_attempts($attemptobj->review_url(null, $page, $showall));
    if ($attemptlist) {
        $summarydata['attemptlist'] = array(
            'title'   => get_string('attempts', 'prova'),
            'content' => $attemptlist,
        );
    }
}

// Timing information.
$summarydata['startedon'] = array(
    'title'   => get_string('startedon', 'prova'),
    'content' => userdate($attempt->timestart),
);

$summarydata['state'] = array(
    'title'   => get_string('attemptstate', 'prova'),
    'content' => prova_attempt::state_name($attempt->state),
);

if ($attempt->state == prova_attempt::FINISHED) {
    $summarydata['completedon'] = array(
        'title'   => get_string('completedon', 'prova'),
        'content' => userdate($attempt->timefinish),
    );
    $summarydata['timetaken'] = array(
        'title'   => get_string('timetaken', 'prova'),
        'content' => $timetaken,
    );
}

if (!empty($overtime)) {
    $summarydata['overdue'] = array(
        'title'   => get_string('overdue', 'prova'),
        'content' => $overtime,
    );
}

// Show marks (if the user is allowed to see marks at the moment).
$grade = prova_rescale_grade($prova->sumgrades, $prova, false);
if ($options->marks >= question_display_options::MARK_AND_MAX && prova_has_grades($prova)) {

    if ($attempt->state == prova_attempt::IN_PROGRESS) {
        // Cannot display grade.

    } else if (!$prova->corrected || is_null($grade)) {
        $summarydata['grade'] = array(
            'title'   => get_string('grade', 'prova'),
            'content' => modprova_format_grade($prova, $grade),
        );

    } else {
        // Show raw marks only if they are different from the grade (like on the view page).
        $a = new stdClass();
        $a->grade = modprova_format_grade($prova, $attempt->sumgrades);
        $a->maxgrade = modprova_format_grade($prova, $prova->sumgrades);
        $summarydata['marks'] = array(
            'title' => '<span class="reviewnota">' . get_string('marks', 'prova') . '<span>',
            'content' => '<span class="reviewnota">' . get_string('outofshort', 'prova', $a) . '<span>',
        );

        // Now the scaled grade.
        // $a = new stdClass();
        // $a->grade = html_writer::tag('b', modprova_format_grade($prova, $grade));
        // $a->maxgrade = modprova_format_grade($prova, $attempt->grade);
        // if ($prova->grade != 100) {
        //     $a->percent = html_writer::tag('b', format_float(
        //             $attempt->sumgrades * 100 / $prova->sumgrades, 0));
        //     $formattedgrade = get_string('outofpercent', 'prova', $a);
        // } else {
        //     $formattedgrade = get_string('outof', 'prova', $a);
        // }
        // $summarydata['grade'] = array(
        //     'title'   => get_string('grade', 'prova'),
        //     'content' => $formattedgrade,
        // );
    }
}

// Any additional summary data from the behaviour.
$summarydata = array_merge($summarydata, $attemptobj->get_additional_summary_data($options));

// Summary table end. ==============================================================================

if ($showall) {
    $slots = $attemptobj->get_slots();
    $lastpage = true;
} else {
    $slots = $attemptobj->get_slots($page);
    $lastpage = $attemptobj->is_last_page($page);
}

$output = $PAGE->get_renderer('mod_prova');

// Arrange for the navigation to be displayed.
// $navbc = $attemptobj->get_navigation_panel($output, 'prova_review_nav_panel', $page, $showall);
// $regions = $PAGE->blocks->get_regions();
// $PAGE->blocks->add_fake_block($navbc, reset($regions));

echo $output->review_page($attemptobj, $slots, $page, $showall, $lastpage, $options, $summarydata, $printter);

// Trigger an event for this review.
$attemptobj->fire_attempt_reviewed_event();
