<?php
/**
 * Helper functions for the prova reports.
 *
 * @package mod_prova
 */

defined('MOODLE_INTERNAL') || die();

require_once($CFG->dirroot . '/mod/prova/lib.php');
require_once($CFG->libdir . '/filelib.php');

/**
 * Returns an array of reports to which the current user has access to.
 * @return array reports are ordered as they should be for display in tabs.
 */
function prova_report_list($context) {
    global $DB;
    static $reportlist = null;
    if (!is_null($reportlist)) {
        return $reportlist;
    }

    $reports = $DB->get_records('prova_reports', null, 'displayorder DESC', 'name, capability');
    $reportdirs = core_component::get_plugin_list('prova');

    // Order the reports tab in descending order of displayorder.
    $reportcaps = array();
    foreach ($reports as $key => $report) {
        if (array_key_exists($report->name, $reportdirs)) {
            $reportcaps[$report->name] = $report->capability;
        }
    }

    // Add any other reports, which are on disc but not in the DB, on the end.
    foreach ($reportdirs as $reportname => $notused) {
        if (!isset($reportcaps[$reportname])) {
            $reportcaps[$reportname] = null;
        }
    }
    $reportlist = array();
    foreach ($reportcaps as $name => $capability) {
        if (empty($capability)) {
            $capability = 'mod/prova:viewreports';
        }
        if (has_capability($capability, $context)) {
            $reportlist[] = $name;
        }
    }
    return $reportlist;
}

/**
 * Takes an array of objects and constructs a multidimensional array keyed by
 * the keys it finds on the object.
 * @param array $datum an array of objects with properties on the object
 * including the keys passed as the next param.
 * @param array $keys Array of strings with the names of the properties on the
 * objects in datum that you want to index the multidimensional array by.
 * @param bool $keysunique If there is not only one object for each
 * combination of keys you are using you should set $keysunique to true.
 * Otherwise all the object will be added to a zero based array. So the array
 * returned will have count($keys) + 1 indexs.
 * @return array multidimensional array properly indexed.
 */
function prova_report_index_by_keys($datum, $keys, $keysunique = true) {
    if (!$datum) {
        return array();
    }
    $key = array_shift($keys);
    $datumkeyed = array();
    foreach ($datum as $data) {
        if ($keys || !$keysunique) {
            $datumkeyed[$data->{$key}][]= $data;
        } else {
            $datumkeyed[$data->{$key}]= $data;
        }
    }
    if ($keys) {
        foreach ($datumkeyed as $datakey => $datakeyed) {
            $datumkeyed[$datakey] = prova_report_index_by_keys($datakeyed, $keys, $keysunique);
        }
    }
    return $datumkeyed;
}

function prova_report_unindex($datum) {
    if (!$datum) {
        return $datum;
    }
    $datumunkeyed = array();
    foreach ($datum as $value) {
        if (is_array($value)) {
            $datumunkeyed = array_merge($datumunkeyed, prova_report_unindex($value));
        } else {
            $datumunkeyed[] = $value;
        }
    }
    return $datumunkeyed;
}

/**
 * Are there any questions in this prova?
 * @param int $provaid the seox_prova id.
 */
function prova_has_questions($provaid) {
    global $DB;
    return $DB->record_exists('seox_prova_slots', array('provaid' => $provaid));
}

/**
 * Get the slots of real questions (not descriptions) in this prova, in order.
 * @param object $prova the seox_prova.
 * @return array of slot => $question object with fields
 *      ->slot, ->id, ->maxmark, ->number, ->length.
 */
function prova_report_get_significant_questions($prova) {
    global $DB;

    $qsbyslot = $DB->get_records_sql("
            SELECT slot.slot,
                   q.id,
                   q.qtype,
                   q.length,
                   slot.maxmark

              FROM {question} q
              JOIN {seox_prova_slots} slot ON slot.questionid = q.id

             WHERE slot.provaid = ?
               AND q.length > 0

          ORDER BY slot.slot", array($prova->provaid));

    $number = 1;
    foreach ($qsbyslot as $question) {
        $question->number = $number;
        $number += $question->length;
        $question->type = $question->qtype;
    }

    return $qsbyslot;
}

/**
 * @param object $prova the prova settings.
 * @return bool whether, for this prova, it is possible to filter attempts to show
 *      only those that gave the final grade.
 */
function prova_report_can_filter_only_graded($prova) {
    return $prova->attempts != 1 && $prova->grademethod != QUIZ_GRADEAVERAGE;
}

/**
 * This is a wrapper for {@link prova_report_grade_method_sql} that takes the whole prova object instead of just the grading method
 * as a param. See definition for {@link prova_report_grade_method_sql} below.
 *
 * @param object $prova
 * @param string $provaattemptsalias sql alias for 'prova_attempts' table
 * @return string sql to test if this is an attempt that will contribute towards the grade of the user
 */
function prova_report_qm_filter_select($prova, $provaattemptsalias = 'provaa') {
    if ($prova->attempts == 1) {
        // This prova only allows one attempt.
        return '';
    }
    return prova_report_grade_method_sql($prova->grademethod, $provaattemptsalias);
}

/**
 * Should the grades be displayed in this report. That depends on the prova
 * display options, and whether the prova is graded.
 * @param object $prova the prova settings.
 * @param context $context the prova context.
 * @return bool
 */
function prova_report_should_show_grades($prova, context $context) {
    if ($prova->timeclose && time() > $prova->timeclose) {
        $when = mod_prova_display_options::AFTER_CLOSE;
    } else {
        $when = mod_prova_display_options::LATER_WHILE_OPEN;
    }
    $reviewoptions = mod_prova_display_options::make_from_prova($prova, $when);

    return prova_has_grades($prova) &&
            ($reviewoptions->marks >= question_display_options::MARK_AND_MAX ||
            has_capability('moodle/grade:viewhidden', $context));
}

/**
 * Create a filename for use when downloading data from a prova report. It is
 * expected that this will be passed to flexible_table::is_downloading, which
 * cleans the filename of bad characters and adds the file extension.
 * @param string $report the type of report.
 * @param string $courseshortname the course shortname.
 * @param string $provaname the prova name.
 * @return string the filename.
 */
function prova_report_download_filename($report, $courseshortname, $provaname) {
    return $courseshortname . '-' . format_string($provaname, true) . '-' . $report;
}

function prova_report_highlighting_grading_method($prova, $qmsubselect, $qmfilter) {
    if ($prova->attempts == 1) {
        return '<p>' . get_string('onlyoneattemptallowed', 'prova_responses') . '</p>';

    } else if (!$qmsubselect) {
        return '<p>' . get_string('allattemptscontributetograde', 'prova_responses') . '</p>';

    } else if ($qmfilter) {
        return '<p>' . get_string('showinggraded', 'prova_responses') . '</p>';

    } else {
        return '<p>' . get_string('showinggradedandungraded', 'prova_responses',
                '<span class="gradedattempt">' . prova_get_grading_option_name($prova->grademethod) .
                '</span>') . '</p>';
    }
}

/* function prova_report_feedback_for_grade($grade, $quizid, $context) {
    global $DB;

    static $feedbackcache = array();

    if (!isset($feedbackcache[$quizid])) {
        $feedbackcache[$quizid] = $DB->get_records('quiz_feedback', array('quizid' => $provaid));
    }

    // With CBM etc, it is possible to get -ve grades, which would then not match
    // any feedback. Therefore, we replace -ve grades with 0.
    $grade = max($grade, 0);

    $feedbacks = $feedbackcache[$provaid];
    $feedbackid = 0;
    $feedbacktext = '';
    $feedbacktextformat = FORMAT_MOODLE;
    foreach ($feedbacks as $feedback) {
        if ($feedback->mingrade <= $grade && $grade < $feedback->maxgrade) {
            $feedbackid = $feedback->id;
            $feedbacktext = $feedback->feedbacktext;
            $feedbacktextformat = $feedback->feedbacktextformat;
            break;
        }
    }

    // Clean the text, ready for display.
    $formatoptions = new stdClass();
    $formatoptions->noclean = true;
    $feedbacktext = file_rewrite_pluginfile_urls($feedbacktext, 'pluginfile.php',
            $context->id, 'mod_prova', 'feedback', $feedbackid);
    $feedbacktext = format_text($feedbacktext, $feedbacktextformat, $formatoptions);

    return $feedbacktext;
} */

/**
 * Get the number of students whose score was in a particular band for this quiz.
 * @param number $bandwidth the width of each band.
 * @param int $bands the number of bands
 * @param int $provaid the prova id.
 * @param \core\dml\sql_join $usersjoins (joins, wheres, params) to get enrolled users
 * @return array band number => number of users with scores in that band.
 */
function prova_report_grade_bands($bandwidth, $bands, $provaid, \core\dml\sql_join $usersjoins = null) {
    global $DB;
    if (!is_int($bands)) {
        debugging('$bands passed to quiz_report_grade_bands must be an integer. (' .
                gettype($bands) . ' passed.)', DEBUG_DEVELOPER);
        $bands = (int) $bands;
    }

    if ($usersjoins && !empty($usersjoins->joins)) {
        $userjoin = "JOIN {user} u ON u.id = pg.userid
                {$usersjoins->joins}";
        $usertest = $usersjoins->wheres;
        $params = $usersjoins->params;
    } else {
        $userjoin = '';
        $usertest = '1=1';
        $params = array();
    }
    $sql = "
SELECT band, COUNT(1)

FROM (
    SELECT FLOOR(pg.grade / :bandwidth) AS band
      FROM {prova_grades} pg
    $userjoin
    WHERE $usertest AND pg.prova = :provaid
) subquery

GROUP BY
    band

ORDER BY
    band";

    $params['provaid'] = $provaid;
    $params['bandwidth'] = $bandwidth;

    $data = $DB->get_records_sql_menu($sql, $params);

    // We need to create array elements with values 0 at indexes where there is no element.
    $data = $data + array_fill(0, $bands + 1, 0);
    ksort($data);

    // Place the maximum (perfect grade) into the last band i.e. make last
    // band for example 9 <= g <=10 (where 10 is the perfect grade) rather than
    // just 9 <= g <10.
    $data[$bands - 1] += $data[$bands];
    unset($data[$bands]);

    return $data;
}