<?php
/**
 * This file defines the prova responses report class.
 *
 * @package prova_responses
 */


defined('MOODLE_INTERNAL') || die();

require_once($CFG->dirroot . '/mod/prova/src/report/attemptsreport.php');
require_once($CFG->dirroot . '/mod/prova/src/report/responses/responses_options.php');
require_once($CFG->dirroot . '/mod/prova/src/report/responses/responses_form.php');
require_once($CFG->dirroot . '/mod/prova/src/report/responses/last_responses_table.php');
require_once($CFG->dirroot . '/mod/prova/src/report/responses/first_or_all_responses_table.php');


/**
 * Quiz report subclass for the responses report.
 *
 * This report lists some combination of
 *  * what question each student saw (this makes sense if random questions were used).
 *  * the response they gave,
 *  * and what the right answer is.
 *
 * Like the overview report, there are options for showing students with/without
 * attempts, and for deleting selected attempts.
 *
 * @copyright 1999 onwards Martin Dougiamas and others {@link http://moodle.com}
 * @license   http://www.gnu.org/copyleft/gpl.html GNU GPL v3 or later
 */
class prova_responses_report extends prova_attempts_report {

    protected $corrected_btn = null;
    protected $table = null;

    public function display($prova, $cm, $course) {
        global $OUTPUT, $DB, $PAGE;

        $PAGE->requires->css('/mod/prova/src/report/responses/status.css');

        list($currentgroup, $studentsjoins, $groupstudentsjoins, $allowedjoins) = $this->init(
                'responses', 'prova_responses_settings_form', $prova, $cm, $course);

        $options = new prova_responses_options('responses', $prova, $cm, $course);

        if ($fromform = $this->form->get_data()) {
            $options->process_settings_from_form($fromform);
        } else {
            $options->process_settings_from_params();
        }

        $this->form->set_data($options->get_initial_form_data());

        // Load the required questions.
        $questions = prova_report_get_significant_questions($prova);

        // Prepare for downloading, if applicable.
        $courseshortname = format_string($course->shortname, true,
                array('context' => context_course::instance($course->id)));
        if ($options->whichtries === question_attempt::LAST_TRY) {
            $tableclassname = 'prova_last_responses_table';
        } else {
            $tableclassname = 'prova_first_or_all_responses_table';
        }
        $table = new $tableclassname($prova, $this->context, $this->qmsubselect,
                $options, $groupstudentsjoins, $studentsjoins, $questions, $options->get_url());
        $filename = prova_report_download_filename(get_string('responsesfilename', 'prova_responses'),
                $courseshortname, $prova->name);
        $table->is_downloading($options->download, $filename,
                $courseshortname . ' ' . format_string($prova->name, true));
        if ($table->is_downloading()) {
            raise_memory_limit(MEMORY_EXTRA);
        }

        $this->hasgroupstudents = false;
        if (!empty($groupstudentsjoins->joins)) {
            $sql = "SELECT DISTINCT u.id
                      FROM {user} u
                    $groupstudentsjoins->joins
                     WHERE $groupstudentsjoins->wheres";
            $this->hasgroupstudents = $DB->record_exists_sql($sql, $groupstudentsjoins->params);
        }
        $hasstudents = false;
        if (!empty($studentsjoins->joins)) {
            $sql = "SELECT DISTINCT u.id
                    FROM {user} u
                    $studentsjoins->joins
                    WHERE $studentsjoins->wheres";
            $hasstudents = $DB->record_exists_sql($sql, $studentsjoins->params);
        }
        if ($options->attempts == self::ALL_WITH) {
            // This option is only available to users who can access all groups in
            // groups mode, so setting allowed to empty (which means all prova attempts
            // are accessible, is not a security problem.
            $allowedjoins = new \core\dml\sql_join();
        }

        $this->process_actions($prova, $cm, $currentgroup, $groupstudentsjoins, $allowedjoins, $options->get_url());

        $hasquestions = prova_has_questions($prova->provaid);

        // Start output.
        if (!$table->is_downloading()) {
            // Only print headers if not asked to download data.
            $this->print_standard_header_and_messages($cm, $course, $prova,
                    $options, $currentgroup, $hasquestions, $hasstudents);

            
        }

        $hasstudents = $hasstudents && (!$currentgroup || $this->hasgroupstudents);
        if ($hasquestions && ($hasstudents || $options->attempts == self::ALL_WITH)) {

            $table->setup_sql_queries($allowedjoins);

            if (!$table->is_downloading()) {
                // Print information on the grading method.
                if ($strattempthighlight = prova_report_highlighting_grading_method(
                        $prova, $this->qmsubselect, $options->onlygraded)) {
                    echo '<div class="provaattemptcounts">' . $strattempthighlight . '</div>';

                    // Print the display options.
                    $this->form->display();
                }
            }

            // Define table columns.
            $columns = array();
            $headers = array();

            if (!$table->is_downloading() && $options->checkboxcolumn) {
                $columns[] = 'checkbox';
                $headers[] = null;
            }

            $this->add_user_columns($table, $columns, $headers);
            $this->add_state_column($columns, $headers);

            if ($table->is_downloading()) {
                $this->add_time_columns($columns, $headers);
            }

            $this->add_grade_columns($prova, $options->usercanseegrades, $columns, $headers);

            if ($options->showreducedmode) {
                $columns[] = 'status' . implode('-',array_keys($questions));
                $headers[] = get_string('reportstatuslabel', 'prova');
            }

            if ($options->showqtext || $options->showresponses || $options->showright) {
                foreach ($questions as $id => $question) {
                    if ($options->showqtext) {
                        $columns[] = 'question' . $id;
                        $headers[] = get_string('questionx', 'question', $question->number);
                    }
                    if ($options->showresponses) {
                        $columns[] = 'response' . $id;
                        $headers[] = get_string('responsex', 'prova_responses', $question->number);
                    }
                    if ($options->showright) {
                        $columns[] = 'right' . $id;
                        $headers[] = get_string('rightanswerx', 'prova_responses', $question->number);
                    }
                }
            }

            $table->define_columns($columns);
            $table->define_headers($headers);
            $table->sortable(true, 'uniqueid');

            // Set up the table.
            $table->define_baseurl($options->get_url());

            $this->configure_user_columns($table);

            $table->no_sorting('feedbacktext');
            $table->column_class('sumgrades', 'bold');

            $table->set_attribute('id', 'responses');

            $table->collapsible(true);

            $table->out($options->pagesize, true);


            if (!$table->is_downloading()) {
                $output = '';
                if (!$prova->corrected) {
                    $provaid = $prova->id;
                    // Finish attempt button.
                    $options = array(
                        'provaid' => $provaid,
                        'sesskey' => sesskey()
                    );

                    $q_diffs = $table->calculate_questions_diff();
                    echo '<script>console.log("report", '.json_encode($q_diffs).')</script>';
                    $options = array_merge($options, $q_diffs);

                    $button = new single_button(new moodle_url('/mod/prova/screens/report/processreport.php', $options), get_string('correctedformbtn', 'prova'));
                    $button->id = 'correctedform';

                    $message = '';
                    if ($table->inprogess_attempts) {
                        $message .= get_string('confirmclosecorrectedform_attemptstate', 'prova', '<strong>'.$table->inprogess_attempts.'</strong>');
                    } 
                    if ($table->notcorrected_attempts) {
                        $message .= (($message != '') ? '<br>' : '') . get_string('confirmclosecorrectedform_correction', 'prova', '<strong>'.$table->notcorrected_attempts.'</strong>');
                    }
                    $message .= (($message != '') ? '<br>' : '') . get_string('confirmclosecorrectedform', 'prova');

                    $button->add_action(new confirm_action($message, null, get_string('submitcorrectedform', 'prova')));

                    $output .= '</form>';
                    $output .= $OUTPUT->render($button);
                    
                } else {
                    $output .= '<span id="corrected-done-alert">' . get_string('submitcorrectedform_alert', 'prova') . '<span>';
                }

                echo $OUTPUT->container($OUTPUT->container($output, 'controls'), 'submitbtns mdl-align');
            }
        }
        return true;
    }

}
