<?php
/**
 * This page is the entry page into the prova UI. Displays information about the
 * prova to students and teachers, and lets students see their previous attempts.
 *
 * @package mod_prova
 */

require_once(__DIR__ . '/../../config.php');
require_once($CFG->libdir.'/gradelib.php');
require_once($CFG->dirroot.'/mod/prova/locallib.php');
require_once($CFG->libdir . '/completionlib.php');
require_once($CFG->dirroot . '/course/format/lib.php');

$id = optional_param('id', 0, PARAM_INT); // Course Module ID, or ...
$q = optional_param('q',  0, PARAM_INT);  // Prova ID.

if ($id) {
    if (!$cm = get_coursemodule_from_id('prova', $id)) {
        print_error('invalidcoursemodule');
    }
    if (!$course = $DB->get_record('course', array('id' => $cm->course))) {
        print_error('coursemisconf');
    }
} else {
    if (!$prova = $DB->get_record('prova', array('id' => $q))) {
        print_error('invalidprovaid', 'prova');
    }
    if (!$course = $DB->get_record('course', array('id' => $prova->course))) {
        print_error('invalidcourseid');
    }
    if (!$cm = get_coursemodule_from_instance("prova", $prova->id, $course->id)) {
        print_error('invalidcoursemodule');
    }
}

// Check login and get context.
require_login($course, false, $cm);
$context = context_module::instance($cm->id);
require_capability('mod/prova:view', $context);

// Cache some other capabilities we use several times.
$canattempt = has_capability('mod/prova:attempt', $context);
$canreviewmine = has_capability('mod/prova:reviewmyattempts', $context);
$canpreview = has_capability('mod/prova:preview', $context);

// Create an object to manage all the other (non-roles) access rules.
$timenow = time();
$provaobj = prova::create($cm->instance, $USER->id);
$accessmanager = new prova_access_manager($provaobj, $timenow, has_capability('mod/prova:ignoretimelimits', $context, null, false));
$prova = $provaobj->get_prova();

// Trigger course_module_viewed event and completion.
prova_view($prova, $course, $cm, $context);

// Initialize $PAGE, compute blocks.
$PAGE->set_url('/mod/prova/view.php', array('id' => $cm->id));

// Create view object which collects all the information the renderer will need.
$viewobj = new mod_prova_view_object();
$viewobj->accessmanager = $accessmanager;
$viewobj->canreviewmine = $canreviewmine || $canpreview;

// Get this user's attempts.
$attempts = prova_get_user_attempts($prova->id, $USER->id, 'finished', true);
$lastfinishedattempt = end($attempts);
$unfinished = false;
$unfinishedattemptid = null;
if ($unfinishedattempt = prova_get_user_attempt_unfinished($prova->id, $USER->id)) {
    $attempts[] = $unfinishedattempt;

    // If the attempt is now overdue, deal with that - and pass isonline = false.
    // We want the student notified in this case.
    $provaobj->create_attempt_object($unfinishedattempt)->handle_if_time_expired(time(), false);

    $unfinished = $unfinishedattempt->state == prova_attempt::IN_PROGRESS ||
            $unfinishedattempt->state == prova_attempt::OVERDUE;
    if (!$unfinished) {
        $lastfinishedattempt = $unfinishedattempt;
    }
    $unfinishedattemptid = $unfinishedattempt->id;
    $unfinishedattempt = null; // To make it clear we do not use this again.
}
$numattempts = count($attempts);

$viewobj->attempts = $attempts;
$viewobj->attemptobjs = array();
foreach ($attempts as $attempt) {
    $viewobj->attemptobjs[] = prova_attempt::create($attempt->id);
}

// Work out the final grade, checking whether it was overridden in the gradebook.
if (!$canpreview) {
    $mygrade = prova_get_best_grade($prova, $USER->id);
} else if ($lastfinishedattempt) {
    // Users who can preview the prova don't get a proper grade, so work out a
    // plausible value to display instead, so the page looks right.
    $mygrade = prova_rescale_grade($lastfinishedattempt->sumgrades, $prova, false);
} else {
    $mygrade = null;
}

$mygradeoverridden = false;
$gradebookfeedback = '';

$grading_info = grade_get_grades($course->id, 'mod', 'prova', $prova->id, $USER->id);
if (!empty($grading_info->items)) {
    $item = $grading_info->items[0];
    if (isset($item->grades[$USER->id])) {
        $grade = $item->grades[$USER->id];

        if ($grade->overridden) {
            $mygrade = $grade->grade + 0; // Convert to number.
            $mygradeoverridden = true;
        }
        if (!empty($grade->str_feedback)) {
            $gradebookfeedback = $grade->str_feedback;
        }
    }
}

$title = $course->shortname . ': ' . format_string($prova->name);
$PAGE->set_title($title);
$PAGE->set_heading($course->fullname);
$output = $PAGE->get_renderer('mod_prova');

// Print table with existing attempts.
if ($attempts) {
    // Work out which columns we need, taking account what data is available in each attempt.
    list($someoptions, $alloptions) = prova_get_combined_reviewoptions($prova, $attempts);

    $viewobj->attemptcolumn  = $prova->attempts != 1;

    $viewobj->gradecolumn    = $someoptions->marks >= question_display_options::MARK_AND_MAX && prova_has_grades($prova);
    $viewobj->markcolumn     = $viewobj->gradecolumn && ($prova->grade != $prova->sumgrades);
    $viewobj->overallstats   = $lastfinishedattempt && $alloptions->marks >= question_display_options::MARK_AND_MAX;

    // TODO: feedback. Depois de implementar correção.
    // $viewobj->feedbackcolumn = prova_has_feedback($prova) && $alloptions->overallfeedback;
}

$viewobj->timenow = $timenow;
$viewobj->numattempts = $numattempts;
$viewobj->mygrade = $mygrade;
$viewobj->moreattempts = $unfinished ||
        !$accessmanager->is_finished($numattempts, $lastfinishedattempt);
$viewobj->mygradeoverridden = $mygradeoverridden;
$viewobj->gradebookfeedback = $gradebookfeedback;
$viewobj->lastfinishedattempt = $lastfinishedattempt;
// TODO: verificar necessidade de edit.php
// $viewobj->canedit = has_capability('mod/prova:manage', $context);
// $viewobj->editurl = new moodle_url('/mod/prova/edit.php', array('cmid' => $cm->id));
$viewobj->backtocourseurl = new moodle_url('/course/view.php', array('id' => $course->id));
$viewobj->startattempturl = $provaobj->start_attempt_url();

if ($accessmanager->is_preflight_check_required($unfinishedattemptid)) {
    $viewobj->preflightcheckform = $accessmanager->get_preflight_check_form(
            $viewobj->startattempturl, $unfinishedattemptid);
}
$viewobj->popuprequired = $accessmanager->attempt_must_be_in_popup();
$viewobj->popupoptions = $accessmanager->get_popup_options();

// Display information about this prova.
$viewobj->infomessages = $viewobj->accessmanager->describe_rules();
if ($prova->attempts != 1) {
    $viewobj->infomessages[] = get_string('gradingmethod', 'prova', prova_get_grading_option_name($prova->grademethod));
}

// Determine wheter a start attempt button should be displayed.
$viewobj->provahasquestions = $provaobj->has_questions();
$viewobj->preventmessages = array();
if (!$viewobj->provahasquestions) {
    $viewobj->buttontext = '';

} else {
    if ($unfinished) {
        if ($canattempt) {
            $viewobj->buttontext = get_string('continueattemptprova', 'prova');
        } else if ($canpreview) {
            $viewobj->buttontext = get_string('continuepreview', 'prova');
        }

    } else {
        if ($canattempt) {
            $viewobj->preventmessages = $viewobj->accessmanager->prevent_new_attempt(
                    $viewobj->numattempts, $viewobj->lastfinishedattempt);
            if ($viewobj->preventmessages) {
                $viewobj->buttontext = '';
            } else if ($viewobj->numattempts == 0) {
                $viewobj->buttontext = get_string('attemptprovanow', 'prova');
            } else {
                $viewobj->buttontext = get_string('reattemptprova', 'prova');
            }

        } else if ($canpreview) {
            $viewobj->buttontext = get_string('previewprovanow', 'prova');
        }
    }

    // If, so far, we think a button should be printed, so check if they will be
    // allowed to access it.
    if ($viewobj->buttontext) {
        if (!$viewobj->moreattempts) {
            $viewobj->buttontext = '';
        } else if ($canattempt
                && $viewobj->preventmessages = $viewobj->accessmanager->prevent_access()) {
            $viewobj->buttontext = '';
        }
    }
}

$viewobj->showbacktocourse = ($viewobj->buttontext === '' &&
        course_get_format($course)->has_view_page());

$viewobj->teacherormanager = has_capability('mod/prova:addinstance', $context);

if($prova->corrected){
    $PAGE->requires->css('/mod/prova/styles.css');
}

echo $OUTPUT->header();

if (isguestuser()) {
    // Guests can't do a prova, so offer them a choice of logging in or going back.
    echo $output->view_page_guest($course, $prova, $cm, $context, $viewobj->infomessages);
} else if (!isguestuser() && !($canattempt || $canpreview
          || $viewobj->canreviewmine)) {
    // If they are not enrolled in this course in a good enough role, tell them to enrol.
    echo $output->view_page_notenrolled($course, $prova, $cm, $context, $viewobj->infomessages);
} else {
    echo $output->view_page($course, $prova, $cm, $context, $viewobj);
}


echo $OUTPUT->footer();
